"""
Additional global constraints which are not (yet) part of the standard CPMpy collection.

This file contains all the missing global constraints in order to support XCSP3-core, 
which is a restricted scope of the complete XCSP3 specification (as used for the competitions).

Currently, version 3.2 is supported.


===============
List of classes
===============

.. autosummary::
    :nosignatures:

    AllDifferentLists
    AllDifferentListsExceptN
    SubCircuit
    SubCircuitWithStart
    SafeOnlyInverse
    InverseOne
    Channel
    NonReifiedTable
    RowSelectingShortTable
    NegativeShortTable
    MDD
    Regular
    NotInDomain
    NoOverlap2d
    IfThenElseNum
    DynamicCumulative

=============================
List of Solver-native classes
=============================

.. autosummary::
    :nosignatures:

    OrtNoOverlap2D
    OrtSubcircuit
    OrtSubcircuitWithStart
    ChocoSubcircuit
    MinizincSubcircuit
    MinizincSubcircuitWithStart
"""

import numpy as np
import cpmpy as cp
from cpmpy import cpm_array, intvar, boolvar
from cpmpy.exceptions import CPMpyException
from cpmpy.expressions.core import Expression, Operator
from cpmpy.expressions.globalconstraints import GlobalConstraint, GlobalFunction, AllDifferent, InDomain, DirectConstraint
from cpmpy.expressions.utils import STAR, is_any_list, is_num, all_pairs, argvals, flatlist, is_boolexpr, argval, is_int, \
    get_bounds, eval_comparison
from cpmpy.expressions.variables import _IntVarImpl


# ---------------------------------------------------------------------------- #
#                                Generic Globals                               #
# ---------------------------------------------------------------------------- #
class AllDifferentLists(GlobalConstraint):
    """
        Ensures none of the lists given are exactly the same.
        Called 'lex_alldifferent' in the global constraint catalog:
        https://sofdem.github.io/gccat/gccat/Clex_alldifferent.html#uid24923
    """
    def __init__(self, lists):
        if any(not is_any_list(lst) for lst in lists):
            raise TypeError(f"AllDifferentLists expects a list of lists, but got {lists}")
        if any(len(lst) != len(lists[0]) for lst in lists):
            raise ValueError("Lists should have equal length, but got these lengths:", list(map(len, lists)))
        super().__init__("alldifferent_lists", [flatlist(lst) for lst in lists])

    def decompose(self):
        """Returns the decomposition
        """
        from cpmpy.expressions.python_builtins import any as cpm_any
        constraints = []
        for lst1, lst2 in all_pairs(self.args):
            constraints += [cpm_any(var1 != var2 for var1, var2 in zip(lst1, lst2))]
        return constraints, []

    def value(self):
        lst_vals = [tuple(argvals(a)) for a in self.args]
        return len(set(lst_vals)) == len(self.args)
class AllDifferentListsExceptN(GlobalConstraint):
    """
        Ensures none of the lists given are exactly the same. Excluding the tuples given in N
        Called 'lex_alldifferent' in the global constraint catalog:
        https://sofdem.github.io/gccat/gccat/Clex_alldifferent.html#uid24923
    """
    def __init__(self, lists, n):
        if not is_any_list(n):
            raise TypeError(f"AllDifferentListsExceptN expects a (list of) lists to exclude but got {n}")
        if any(not is_any_list(x) for x in n): #only one list given, not a list of lists
            n = [n]
        for lst in n:
            if not all(is_num(x) for x in lst):
                raise TypeError("Can only use constants as excepting argument")
        if any(not is_any_list(lst) for lst in lists):
            raise TypeError(f"AllDifferentListsExceptN expects a list of lists, but got {lists}")
        if any(len(lst) != len(lists[0]) for lst in lists + n):
            raise ValueError("Lists should have equal length, but got these lengths:", list(map(len, lists)))
        super().__init__("alldifferent_lists_except_n", [[flatlist(lst) for lst in lists], [flatlist(x) for x in n]])

    def decompose(self):
        """Returns the decomposition
        """
        from cpmpy.expressions.python_builtins import all as cpm_all
        constraints = []
        for lst1, lst2 in all_pairs(self.args[0]):
            constraints += [cpm_all(var1 == var2 for var1, var2 in zip(lst1, lst2)).implies(NonReifiedTable(lst1, self.args[1]))]
        return constraints, []

    def value(self):
        lst_vals = [tuple(argvals(a)) for a in self.args[0]]
        except_vals = [tuple(argvals(a)) for a in self.args[1]]
        return len(set(lst_vals) - set(except_vals)) == len([x for x in lst_vals if x not in except_vals])

class SubCircuit(GlobalConstraint):
    """
        The sequence of variables form a subcircuit, where x[i] = j means that j is the successor of i.
        Contrary to Circuit, there is no requirement on all nodes needing to be part of the circuit.
        Nodes which aren't part of the subcircuit, should self loop i.e. x[i] = i.
        The subcircuit can be empty (all stops self-loop).
        A length 1 subcircuit is treated as an empty subcircuit.
        Global Constraint Catalog:
        https://sofdem.github.io/gccat/gccat/Cproper_circuit.html
    """

    def __init__(self, *args):
        flatargs = flatlist(args)

        # Ensure all args are integer successor values
        if any(is_boolexpr(arg) for arg in flatargs):
            raise TypeError("SubCircuit global constraint only takes arithmetic arguments: {}".format(flatargs))
        # Ensure there are at least two stops to create a circuit with
        if len(flatargs) < 2:
            raise CPMpyException("SubCircuitWithStart constraint must be given a minimum of 2 variables for field 'args' as stops to route between.")

        # Create the object
        super().__init__("subcircuit", flatargs)

    def decompose(self):
        """
            Decomposition for SubCircuit
            A mix of the above Circuit decomposition, with elements from the Minizinc implementation for the support of optional visits:
            https://github.com/MiniZinc/minizinc-old/blob/master/lib/minizinc/std/subcircuit.mzn
        """
        from cpmpy.expressions.python_builtins import min as cpm_min
        from cpmpy.expressions.python_builtins import all as cpm_all

        # Input arguments
        succ = cpm_array(self.args) # Successor variables
        n = len(succ)

        # Decision variables
        start_node = intvar(0, n-1) # The first stop in the subcircuit.
        end_node = intvar(0, n-1) # The last stop in the subcircuit, before looping back to the "start_node".
        index_within_subcircuit = intvar(0, n-1, shape=n) # The position each stop takes within the subcircuit, with the assumption that the stop "start_node" gets index 0.
        is_part_of_circuit = boolvar(shape=n) # Whether a stop is part of the subcircuit.
        empty = boolvar() # To detect when a subcircuit is completely empty

        # Constraining
        constraining = []
        constraining += [AllDifferent(succ)] # All stops should have a unique successor.
        constraining += list( is_part_of_circuit.implies(succ < len(succ)) ) # Successor values should remain within domain.
        for i in range(0, n):
            # If a stop is on the subcircuit and it is not the last one, than its successor should have +1 as index.
            constraining += [(is_part_of_circuit[i] & (i != end_node)).implies(
                index_within_subcircuit[succ[i]] == (index_within_subcircuit[i] + 1)
            )]
        constraining += list( is_part_of_circuit == (succ != np.arange(n)) ) # When a node is part of the subcircuit it should not self loop, if it is not part it should self loop.

        # Defining
        defining = []
        defining += [ empty == cpm_all(succ == np.arange(n)) ] # Definition of empty subcircuit (all nodes self-loop)
        defining += [ empty.implies(cpm_all(index_within_subcircuit == cpm_array([0]*n))) ] # If the subcircuit is empty, default all index values to 0
        defining += [ empty.implies(start_node == 0) ] # If the subcircuit is empty, any node could be a start of a 0-length circuit. Default to node 0 as symmetry breaking.
        defining += [succ[end_node] == start_node] # Definition of the last node. As the successor we should cycle back to the start.
        defining += [ index_within_subcircuit[start_node] == 0 ] # The ordering starts at the start_node.
        defining += [ ( empty | (is_part_of_circuit[start_node] == True) ) ] # The start node can only NOT belong to the subcircuit when the subcircuit is empty.
        # Nodes which are not part of the subcircuit get an index fixed to +1 the index of "end_node", which equals the length of the subcircuit.
        # Nodes part of the subcircuit must have an index <= index_within_subcircuit[end_node].
        # The case of an empty subcircuit is an exception, since "end_node" itself is not part of the subcircuit
        defining += [ (is_part_of_circuit[i] == ((~empty) & (index_within_subcircuit[end_node] + 1 != index_within_subcircuit[i]))) for i in range(n)]
        # In a subcircuit any of the visited nodes can be the "start node", resulting in symmetrical solutions -> Symmetry breaking
        # Part of the formulation from the following is used: https://sofdem.github.io/gccat/gccat/Ccycle.html#uid18336
        subcircuit_visits = intvar(0, n-1, shape=n) # The visited nodes in sequence of length n, with possible repeated stops. e.g. subcircuit [0, 2, 1] -> [0, 2, 1, 0, 2, 1]
        defining += [subcircuit_visits[0] == start_node] # The start nodes is the first stop
        defining += [subcircuit_visits[i+1] == succ[subcircuit_visits[i]] for i in range(n-1)] # We follow the successor values
        # The free "start_node" could be any of the values of aux_subcircuit_visits (the actually visited nodes), resulting in degenerate solutions.
        # By enforcing "start_node" to take the smallest value, symmetry breaking is ensured.
        defining += [start_node == cpm_min(subcircuit_visits)]

        return constraining, defining

    def value(self):

        succ = [argval(a) for a in self.args]
        n = len(succ)

        # Find a start_index
        start_index = None
        for i,s in enumerate(succ):
            if i != s:
                # first non self-loop found is taken as start
                start_index = i
                break
        # No valid start found, thus empty subcircuit
        if start_index is None:
            return True # Change to False if empty subcircuits not allowed

        # Check AllDiff
        if not AllDifferent(s).value():
            return False

        # Collect subcircuit
        visited = set([start_index])
        idx = succ[start_index]
        for i in range(len(succ)):
            if idx ==start_index:
                break
            else:
                if idx in visited:
                    return False
            # Check bounds on successor value
            if not (0 <= idx < n): return False
            # Collect
            visited.add(idx)
            idx = succ[idx]

        # Check subcircuit
        for i in range(n):
            # A stop is either visited or self-loops
            if not ( (i in visited) or (succ[i] == i) ):
                return False

        # Check that subcircuit has length of at least 1.
        return succ[start_index] != start_index

class SubCircuitWithStart(GlobalConstraint):

    """
        The sequence of variables form a subcircuit, where x[i] = j means that j is the successor of i.
        Contrary to Circuit, there is no requirement on all nodes needing to be part of the circuit.
        Nodes which aren't part of the subcircuit, should self loop i.e. x[i] = i.
        The size of the subcircuit should be strictly greater than 1, so not all stops can self loop
        (as otherwise the start_index will never get visited).
        start_index will be treated as the start of the subcircuit.
        The only impact of start_index is that it will be guaranteed to be inside the subcircuit.
        Global Constraint Catalog:
        https://sofdem.github.io/gccat/gccat/Cproper_circuit.html
    """

    def __init__(self, *args, start_index:int=0):
        flatargs = flatlist(args)

        # Ensure all args are integer successor values
        if any(is_boolexpr(arg) for arg in flatargs):
            raise TypeError("SubCircuitWithStart global constraint only takes arithmetic arguments: {}".format(flatargs))
        # Ensure start_index is an integer
        if not isinstance(start_index, int):
            raise TypeError("SubCircuitWithStart global constraint's start_index argument must be an integer: {}".format(start_index))
        # Ensure that the start_index is within range
        if not ((start_index >= 0) and (start_index < len(flatargs))):
            raise ValueError("SubCircuitWithStart's start_index must be within the range [0, #stops-1] and thus refer to an actual stop as provided through 'args'.")
        # Ensure there are at least two stops to create a circuit with
        if len(flatargs) < 2:
            raise CPMpyException("SubCircuitWithStart constraint must be given a minimum of 2 variables for field 'args' as stops to route between.")

        # Create the object
        super().__init__("subcircuitwithstart", flatargs + [start_index])

    def decompose(self):
        """
            Decomposition for SubCircuitWithStart.
            SubCircuitWithStart simply gets decomposed into SubCircuit and a constraint
            enforcing the start_index to be part of the subcircuit.
        """
        # Get the arguments
        start_index = self.args[-1]
        succ = cpm_array(self.args[:-1]) # Successor variables

        constraining = []
        constraining += [SubCircuit(succ)] # The successor variables should form a subcircuit.
        constraining += [succ[start_index] != start_index] # The start_index should be inside the subcircuit.

        defining = []

        return constraining, defining

    def value(self):
        start_index = self.args[-1]
        succ = [argval(a) for a in self.args[:-1]] # Successor variables

        # Check if we have a valid subcircuit and that the start_index is part of it.
        return SubCircuit(succ).value() and (succ[start_index] != start_index)

class SafeOnlyInverse(GlobalConstraint):
    """
       Inverse (aka channeling / assignment) constraint. 'fwd' and
       'rev' represent inverse functions; that is,

            The symmetric version (where len(fwd) == len(rev)) is defined as:
                fwd[i] == x  <==>  rev[x] == i
            The asymmetric version (where len(fwd) < len(rev)) is defined as:
                fwd[i] == x   =>   rev[x] == i

        The included decomposition only covers the "safe" variant, meaning that the expressions in fwd are assumed to have a domain within [0, len(rev)[.
        To also handle the "unsafe" variant, have a look at :class:`globalconstraints.Inverse <cpmpy.expressions.globalconstraints.Inverse>`.

    """
    def __init__(self, fwd, rev):
        flatargs = flatlist([fwd, rev])
        if any(is_boolexpr(arg) for arg in flatargs):
            raise TypeError("Only integer arguments allowed for global constraint Inverse: {}".format(flatargs))
        if len(fwd) > len(rev):
            raise TypeError("len(fwd) should be equal to len(rev) for the symmetric inverse, or smaller than len(rev) for the asymmetric inverse")
        if len(fwd) == len(rev):
            name = "inverse"
        else:
            name = "inverseAsym"
        super().__init__(name, [fwd, rev])

    def decompose(self):
        fwd, rev = self.args
        rev = cpm_array(rev)
        return [cp.all(rev[x] == i for i, x in enumerate(fwd))], []

    def value(self):
        fwd = argvals(self.args[0])
        rev = argvals(self.args[1])
        # args are fine, now evaluate actual inverse cons
        try:
            return all(rev[x] == i for i, x in enumerate(fwd))
        except IndexError: # partiality of Element constraint
            return False

class InverseOne(GlobalConstraint):
    """
       Inverse (aka channeling / assignment) constraint but with only one array.
       Equivalent to Inverse(x,x)
                arr[i] == j  <==>  arr[j] == i
    """
    def __init__(self, arr):
        flatargs = flatlist([arr])
        if any(is_boolexpr(arg) for arg in flatargs):
            raise TypeError("Only integer arguments allowed for global constraint Inverse: {}".format(flatargs))
        super().__init__("inverseOne", [arr])

    def decompose(self):
        from cpmpy.expressions.python_builtins import all
        arr = self.args[0]
        arr = cpm_array(arr)
        return [all(arr[x] == i for i, x in enumerate(arr))], []

    def value(self):
        valsx = argvals(self.args[0])
        try:
            return all(valsx[x] == i for i, x in enumerate(valsx))
        except IndexError: # partiality of Element constraint
            return False


class Channel(GlobalConstraint):
    """
        Channeling constraint. Channeling integer representation of a variable into a representation with boolean
        indicators
            for all 0<=i<len(arr) : arr[i] = 1 <=> value = i
            exists 0<=i<len(arr) s.t. arr[i] = 1
    """
    def __init__(self, arr, v):
        flatargs = flatlist([arr])
        if not all(x.lb >= 0 and x.ub <= 1 for x in flatargs):
            raise TypeError(
                "the first argument of a Channel constraint should only contain 0-1 variables/expressions (i.e., " +
                "intvars/intexprs with domain {0,1} or boolvars/boolexprs)")
        super().__init__("channelValue", [arr, v])

    def decompose(self):
        arr, v = self.args
        return [(arr[i] == 1) == (v == i) for i in range(len(arr))] + [v >= 0, v < len(arr)], []

    def value(self):
        arr, v = self.args
        return sum(argvals(x) for x in arr) == 1 and 0 <= argval(v) < len(arr) and arr[argval(v)] == 1


class NonReifiedTable(GlobalConstraint):
    """
    The values of the variables in 'array' correspond to a row in 'table'.
    This global represents the non-reified version, meaning that it does not support occuring reified when decomposing.
    Look at :class:`globalconstraints.Table <cpmpy.expressions.globalconstraints.Table` for a formulation that does support reification.
    """

    def __init__(self, array, table):
        array = flatlist(array)
        if not all(isinstance(x, Expression) for x in array):
            raise TypeError("the first argument of a Table constraint should only contain variables/expressions")
        super().__init__("table", [array, table])

    def decompose(self):
        """
            This decomposition is only valid in a non-reified setting.
        """
        from cpmpy.expressions.python_builtins import any, all
        arr, tab = self.args
        row_selected = boolvar(shape=len(tab))
        if len(tab) == 1:
            return [all(t == a for (t, a) in zip(tab[0], arr))], []
        
        cons = []
        for i, row in enumerate(tab):
            subexpr = Operator("and", [x == v for x,v in zip(arr, row)])
            cons.append(Operator("->", [row_selected[i], subexpr]))

        return [Operator("or", row_selected)]+cons,[]

    def value(self):
        arr, tab = self.args
        arrval = argvals(arr)
        return arrval in tab

    @property
    def vars(self):
        return self._args[0]

    # specialisation to avoid recursing over big tables
    def has_subexpr(self):
        if not hasattr(self, '_has_subexpr'):  # if _has_subexpr has not been computed before or has been reset
            arr, tab = self.args  # the table 'tab' can only hold constants, never a nested expression
            self._has_subexpr = any(a.has_subexpr() for a in arr)
        return self._has_subexpr

class RowSelectingShortTable(GlobalConstraint):
    """
        Extension of the `Table` constraint where the `table` matrix may contain wildcards (STAR), meaning there are
        no restrictions for the corresponding variable in that tuple.
        This global skips typechecks on the table and has a different decomposition than the standard 
        :class:`globalconstraints.ShortTable <cpmpy.expressions.globalconstraints.ShortTable>`.
    """
    def __init__(self, array, table):
        array = flatlist(array)
        if not all(isinstance(x, Expression) for x in array):
            raise TypeError("The first argument of a Table constraint should only contain variables/expressions")
        if isinstance(table, np.ndarray): # Ensure it is a list
            table = table.tolist()
        super().__init__("short_table", [array, table])

    def decompose(self):
        """
        Alternative decomposition, similar to `element` from Gleb's paper: "Improved Linearization of Constraint
        Programming Models"
        """
        from cpmpy.expressions.python_builtins import any, all

        arr, tab = self.args
        row_selected = boolvar(shape=(len(tab),))

        cons = []
        for i, row in enumerate(tab):
            subexpr = Operator("and", [ai == ri for ai, ri in zip(arr, row) if ri != STAR])
            cons.append(row_selected[i].implies(subexpr))

        return [any(row_selected)]+cons,[]

    def value(self):
        arr, tab = self.args
        tab = np.array(tab)
        arrval = np.array(argvals(arr))
        for row in tab:
            num_row = row[row != STAR].astype(int)
            num_vals = arrval[row != STAR].astype(int)
            if (num_row == num_vals).all():
                return True
        return False

class NegativeShortTable(GlobalConstraint):
    """The values of the variables in 'array' do not correspond to any row in 'table'
    """
    def __init__(self, array, table):
        array = flatlist(array)
        if not all(isinstance(x, Expression) for x in array):
            raise TypeError("the first argument of a Table constraint should only contain variables/expressions")
        super().__init__("negative_shorttable", [array, table])

    def decompose(self):
        from cpmpy.expressions.python_builtins import all as cpm_all
        from cpmpy.expressions.python_builtins import any as cpm_any
        arr, tab = self.args
        return [cpm_all(cpm_any(ai != ri for ai, ri in zip(arr, row) if ri != "*") for row in tab)], []

    def value(self):
        arr, tab = self.args
        arrval = [argval(a) for a in arr]
        for tup in tab:
            thistup = True
            for aval, tval in zip(arrval, tup):
                if tval != '*':
                    if aval != tval:
                        thistup = False
                        break
            if thistup:
                # found tuple that matches
                return False
        # didn't find tuple that matches
        return True


class MDD(GlobalConstraint):
    """
    MDD-constraint: an MDD (Multi-valued Decision Diagram) is an acyclic layerd graph starting from a single node and
    ending in one. Each edge layer corresponds to a variables and each path corresponds to a solution
    The values of the variables in 'array' correspond to a path in the mdd formed by the transitions in 'transitions'.
    Root node is the first node used as a start in the first transition (i.e. transitions[0][0])

    Arguments:
        array: an array of CPMpy expressions (integer variable, global functions,...)
        transitions: an array of tuples (nodeID, int, nodeID) where nodeID is some unique identifiers for the nodes
            (int or str are fine)
    
    Example:
        The following transitions depict a 3 layer MDD, starting at 'r' and ending in 't'
        ("r", 0, "n1"), ("r", 1, "n2"), ("r", 2, "n3"), ("n1", 2, "n4"), ("n2", 2, "n4"), ("n3", 0, "n5"),
        ("n4", 0, "t"), ("n5", 1, "t")
        Its graphical representation is:

        .. code-block:: text
            
                  r
              0/ |1  \\2     X
            n1   n2   n3
            2| /2    /O     Y
             n4     n5
              0\   /1       Z
                 t
        
        It has 3 paths, corresponding to 3 solution for (X,Y,Z): (0,2,0), (1,2,0) and (2,0,1)
    """

    def __init__(self, array, transitions):
        array = flatlist(array)
        if not all(isinstance(x, Expression) for x in array):
            raise TypeError("The first argument of an MDD constraint should only contain variables/expressions")
        if not all(is_transition(transition) for transition in transitions):
            raise TypeError("The second argument of an MDD constraint should be collection of transitions")
        super().__init__("mdd", [array, transitions])
        self.root_node = transitions[0][0]
        self.mapping = {}
        for s, v, e in transitions:
            self.mapping[(s, v)] = e

    def _transition_to_layer_representation(self):
        """ auxiliary function to compute which nodes belongs to which node-layer and which transition belongs to which
        edge-layer of the MDD, needed to compute decomposition
        """
        arr, transitions = self.args
        nodes_by_level = [[self.root_node]]
        transitions_by_level = []
        tran = transitions
        for i in range(len(arr)): # go through each layer
            nodes_by_level.append([])
            transitions_by_level.append([])
            remaining_tran = []
            for t in tran: # test each transition
                ns, _, ne = t
                if ns in nodes_by_level[i]: # add to the current layer if start node belongs to the node-layer
                    if ne not in nodes_by_level[i + 1]:
                        nodes_by_level[i + 1].append(ne)
                    transitions_by_level[i].append(t)
                else:
                    remaining_tran.append(t)
            tran = remaining_tran
        return nodes_by_level, transitions_by_level

    # auxillary method to transform into layered representation (gather all the node by node-layers)
    def _normalize_layer_representation(self, nodes_by_level, transitions_by_level):
        """ auxiliary function to normalize the names of the nodes in layer by layer representation. Node ID in
        normalized representation goes from 0 to n-1 for each layer. Used by the decomposition of the constraint.
        """
        nb_nodes_by_level = [len(x) for x in nodes_by_level]
        num_mapping = {}
        for lvl in nodes_by_level:
            for i in range(len(lvl)):
                num_mapping[lvl[i]] = i
        transitions_by_level_normalized = [[[num_mapping[n_in], v, num_mapping[n_out]]
                                            for n_in, v, n_out in lvl]
                                           for lvl in transitions_by_level]
        return nb_nodes_by_level, num_mapping, transitions_by_level_normalized


    def decompose(self):
        # Table decomposition (not by decomposition of the mdd into one big table, but by having transitions tables for
        # each layer and auxiliary variables for the nodes. Similar to decomposition of regular into table,
        # but with one table for each layer
        arr, _ = self.args
        lb = [x.lb for x in arr]
        ub = [x.ub for x in arr]
        # transform to layer representation
        nbl, tbl = self._transition_to_layer_representation()
        # normalize the naming of the nodes so it can be use as value for aux variables
        nb_nodes_by_level, num_mapping, transitions_by_level_normalized = self._normalize_layer_representation(nbl, tbl)
        # choose the best decomposition depending on number of levels
        if len(transitions_by_level_normalized) > 2:
            # decomposition with multiple transitions table and aux variables for the nodes
            aux = [intvar(0, nb_nodes) for nb_nodes in nb_nodes_by_level[1:]]
            # complete the MDD with additional dummy transitions to get the false end node also represented,
            # needed so the negation works.
            # I.E., now any assignment have a path in the MDD, some, the solutions, ending in an accepting state
            # (end node of the initial MDD), other, the non-solutions, ending in a rejecting state (dummy end node)
            for i in range(len(arr)):
                # add for each state the missing transition to a dummy node on the next level
                transition_dummy = [[num_mapping[n], v, nb_nodes_by_level[i+1]] for n in nbl[i] for v in range(lb[i], ub[i] + 1) if
                            (n, v) not in self.mapping]
                if i != 0:
                    # add transition from one dummy node to the other (not needed for initial layer as no dummy there)
                    transition_dummy += [[nb_nodes_by_level[i], v, nb_nodes_by_level[i+1]] for v in range(lb[i], ub[i] + 1)]
                # add the new transitions
                transitions_by_level_normalized[i] = transitions_by_level_normalized[i] + transition_dummy
            # optimization for first level (only one node, allows to deal with smaller table on first layer)
            tab_first = [x[1:] for x in transitions_by_level_normalized[0]]
            # defining constraints: aux and arr variables define a path in the augmented-with-negative-path-MDD
            defining = [NonReifiedTable([arr[0], aux[0]], tab_first)] \
                   + [NonReifiedTable([aux[i - 1], arr[i], aux[i]], transitions_by_level_normalized[i]) for i in
                      range(1, len(arr))]
            # constraining constraint: end of the path in accepting node
            constraining = [aux[-1] == 0]
            return constraining, defining
        elif len(transitions_by_level_normalized) == 2:
            # decomposition by unfolding into a table (i.e., extract all paths and list them as table entries),
            # avoid auxiliary variables
            tab = [[t_a[1], t_b[1]] for t_a in transitions_by_level_normalized[0] for t_b in
                   transitions_by_level_normalized[1] if t_a[2] == t_b[0]]
            return [NonReifiedTable(arr, tab)], []

        elif len(transitions_by_level_normalized) == 1:
            # decomposition to inDomain, avoid auxiliary variables and tables
            return [InDomain(arr[0], [t[1] for t in transitions_by_level_normalized[0]])], []

    def value(self):
        arr, transitions = self.args
        arrval = [argval(a) for a in arr]
        curr_node = self.root_node
        for v in arrval:
            if (curr_node, v) in self.mapping:
                curr_node = self.mapping[curr_node, v]
            else:
                return False
        return True # can only have reached end node

class Regular(GlobalConstraint):
    """
    Regular-constraint (or Automaton-constraint)
    Takes as input a sequence of variables and a automaton representation using a transition table.
    The constraint is satisfied if the sequence of variables corresponds to an accepting path in the automaton.

    The automaton is defined by a list of transitions, a starting node and a list of accepting nodes.
    The transitions are represented as a list of tuples, where each tuple is of the form (id1, value, id2).
    An id is an integer or string representing a state in the automaton, and value is an integer representing the value of the variable in the sequence.
    The starting node is an integer or string representing the starting state of the automaton.
    The accepting nodes are a list of integers or strings representing the accepting states of the automaton.

    Example: 
        an automaton that accepts the language 0*10* (exactly 1 variable taking value 1) is defined as:

        .. code-block:: python

            cp.Regular(array = cp.intvar(0,1, shape=4),
                    transitions = [("A",0,"A"), ("A",1,"B"), ("B",0,"C"), ("C",0,"C")],
                    start = "A",
                    accepting = ["C"])
    """
    def __init__(self, array, transitions, start, accepting):
        array = flatlist(array)
        # skip all typechecks for comp
        # if not all(isinstance(x, Expression) for x in array):
        #     raise TypeError("The first argument of a regular constraint should only contain variables/expressions")
        
        # if not is_any_list(transitions):
        #     raise TypeError("The second argument of a regular constraint should be a list of transitions")
        # _node_type = type(transitions[0][0])
        # for s,v,e in transitions:
        #     if not isinstance(s, _node_type) or not isinstance(e, _node_type) or not isinstance(v, int):
        #         raise TypeError(f"The second argument of a regular constraint should be a list of transitions ({_node_type}, int, {_node_type})")
        # if not isinstance(start, _node_type):
        #     raise TypeError("The third argument of a regular constraint should be a node id")
        # if not (is_any_list(accepting) and all(isinstance(e, _node_type) for e in accepting)):
        #     raise TypeError("The fourth argument of a regular constraint should be a list of node ids")
        super().__init__("regular", [array, transitions, start, list(accepting)])

        self.nodes = set()
        self.trans_dict = {}
        for s, v, e in transitions:
            self.nodes.update([s,e])
            self.trans_dict[(s, v)] = e
        self.nodes = sorted(self.nodes)
        # normalize node_ids to be 0..n-1, allows for smaller domains
        self.node_map = {n: i for i, n in enumerate(self.nodes)}

    def decompose(self):
        # Decompose to transition table using Table constraints
        
        arr, transitions, start, accepting = self.args
        lbs, ubs = get_bounds(arr)
        lb, ub = min(lbs), max(ubs)
        
        transitions = [[self.node_map[n_in], v, self.node_map[n_out]] for n_in, v, n_out in transitions]

        # add a sink node for transitions that are not defined
        # --> not necessary for comp, because positive context
        # sink = len(self.nodes)
        # transitions += [[self.node_map[n], v, sink] for n in self.nodes for v in range(lb, ub + 1) if (n, v) not in self.trans_dict]
        # transitions += [[sink, v, sink] for v in range(lb, ub + 1)]

        # keep track of current state when traversing the array
        state_vars = intvar(0, len(self.nodes)-1, shape=len(arr))
        id_start = self.node_map[start]
        # optimization: we know the entry node of the automaton, results in smaller table
        defining = [NonReifiedTable([arr[0], state_vars[0]], [[v,e] for s,v,e in transitions if s == id_start])]        
        # define the rest of the automaton using transition table
        defining += [NonReifiedTable([state_vars[i - 1], arr[i], state_vars[i]], transitions) for i in range(1, len(arr))]
        
        # constraint is satisfied iff last state is accepting
        return [InDomain(state_vars[-1], [self.node_map[e] for e in accepting])], defining

    def value(self):
        arr, transitions, start, accepting = self.args
        arrval = [argval(a) for a in arr]
        curr_node = start
        for v in arrval:
            if (curr_node, v) in self.trans_dict:
                curr_node = self.trans_dict[curr_node, v]
            else:
                return False
        return curr_node in accepting



class NotInDomain(GlobalConstraint):
    """
        The "NotInDomain" constraint, defining non-interval domains for an expression
    """

    def __init__(self, expr, arr):
        super().__init__("NotInDomain", [expr, arr])

    def decompose(self):
        """
        This decomp only works in positive context
        """
        from cpmpy.expressions.python_builtins import any, all
        expr, arr = self.args
        lb, ub = expr.get_bounds()

        defining = []
        #if expr is not a var
        if not isinstance(expr, _IntVarImpl):
            aux = intvar(lb, ub)
            defining.append(aux == expr)
            expr = aux

        if not any(isinstance(a, Expression) for a in arr):
            given = len(set(arr))
            missing = ub + 1 - lb - given
            if missing < 2 * given:  # != leads to double the amount of constraints
                # use == if there is less than twice as many gaps in the domain.
                row_selected = boolvar(shape=missing)
                return [any(row_selected)] + [rs.implies(expr == val) for val,rs in zip(range(lb, ub + 1), row_selected) if val not in arr], defining
        return [all([(expr != a) for a in arr])], defining


    def value(self):
        return argval(self.args[0]) not in argvals(self.args[1])

    def __repr__(self):
        return "{} not in {}".format(self.args[0], self.args[1])


class NoOverlap2d(GlobalConstraint):
    """
        2D-version of the NoOverlap constraint.
        Ensures a set of rectangles is placed on a grid such that they do not overlap.
    """
    def __init__(self, start_x, dur_x, end_x,  start_y, dur_y, end_y):
        assert len(start_x) == len(dur_x) == len(end_x) == len(start_y) == len(dur_y) == len(end_y)
        super().__init__("no_overlap2d", [start_x, dur_x, end_x,  start_y, dur_y, end_y])

    def decompose(self):
        from cpmpy.expressions.python_builtins import any as cpm_any

        start_x, dur_x, end_x,  start_y, dur_y, end_y = self.args
        n = len(start_x)
        cons =  [s + d == e for s,d,e in zip(start_x, dur_x, end_x)]
        cons += [s + d == e for s,d,e in zip(start_y, dur_y, end_y)]

        for i,j in all_pairs(list(range(n))):
            cons += [cpm_any([end_x[i] <= start_x[j], end_x[j] <= start_x[i],
                              end_y[i] <= start_y[j], end_y[j] <= start_y[i]])]
        return cons,[]
    def value(self):
        start_x, dur_x, end_x,  start_y, dur_y, end_y = argvals(self.args)
        n = len(start_x)
        if any(s + d != e for s, d, e in zip(start_x, dur_x, end_x)):
            return False
        if any(s + d != e for s, d, e in zip(start_y, dur_y, end_y)):
            return False
        for i,j in all_pairs(list(range(n))):
            if  end_x[i] > start_x[j] and end_x[j] > start_x[i] and \
                 end_y[i] > start_y[j] and end_y[j] > start_y[i]:
                return False
        return True


class IfThenElseNum(GlobalFunction):
    """
        Function returning x if b is True and otherwise y
    """
    def __init__(self, b, x,y):
        super().__init__("IfThenElseNum",[b,x,y])

    def decompose_comparison(self, cmp_op, cpm_rhs):
        b,x,y = self.args

        lbx,ubx = get_bounds(x)
        lby,uby = get_bounds(y)
        iv = intvar(min(lbx,lby), max(ubx,uby))
        defining = [b.implies(x == iv), (~b).implies(y == iv)]

        return [eval_comparison(cmp_op, iv, cpm_rhs)], defining

    def get_bounds(self):
        b,x,y = self.args
        lbs,ubs = get_bounds([x,y])
        return min(lbs), max(ubs)
    def value(self):
        b,x,y = self.args
        if argval(b):
            return argval(x)
        else:
            return argval(y)


class DynamicCumulative(GlobalConstraint):
    """
        Global cumulative constraint. Used for resource aware scheduling. Ensures that the capacity of the resource is never exceeded.
        When decomposed, dynamically switches between time-resource and task-resource decompositions depending on size of domains.
        Equivalent to :class:`~cpmpy.expressions.globalconstraints.NoOverlap` when demand and capacity are equal to 1.
        Supports both varying demand across tasks or equal demand for all jobs.
    """
    def __init__(self, start, duration, end, demand, capacity):
        assert is_any_list(start), "start should be a list"
        assert is_any_list(duration), "duration should be a list"
        assert is_any_list(end), "end should be a list"

        start = flatlist(start)
        duration = flatlist(duration)
        end = flatlist(end)
        assert len(start) == len(duration) == len(end), "Start, duration and end should have equal length"
        n_jobs = len(start)

        for lb in get_bounds(duration)[0]:
            if lb < 0:
                raise TypeError("Durations should be non-negative")

        if is_any_list(demand):
            demand = flatlist(demand)
            assert len(demand) == n_jobs, "Demand should be supplied for each task or be single constant"
        else: # constant demand
            demand = [demand] * n_jobs

        super(DynamicCumulative, self).__init__("cumulative", [start, duration, end, demand, capacity])

    def decompose(self):
        """
            Decomposition from:
            Schutt, Andreas, et al. "Why cumulative decomposition is not as bad as it sounds."
            International Conference on Principles and Practice of Constraint Programming. Springer, Berlin, Heidelberg, 2009.

            Heuristically switches between time-resource and task-resource decomposition depending on the relative size of the time horizon and the number of tasks.
            
            If 
                n = number of tasks
                t = size of time horizon
            then    
                time-resource decomposition scales with n*t 
                task-resource decomposition scales with 3n(n-1)
            thus
                switch when t > 3*n
        """

        arr_args = (cpm_array(arg) if is_any_list(arg) else arg for arg in self.args)
        start, duration, end, demand, capacity = arr_args

        num_tasks = len(demand) # number of tasks
        lb, ub = min(get_bounds(start)[0]), max(get_bounds(end)[1])
        time_horizon =  ub - lb

        cons = []

        if time_horizon > 3 * num_tasks:
            version = "task"
        else:
            version = "time"

        if version == "time":
            # set duration of tasks
            for t in range(len(start)):
                cons += [start[t] + duration[t] == end[t]]

            # demand doesn't exceed capacity
            for t in range(lb,ub+1):
                demand_at_t = 0
                for job in range(len(start)):
                    if is_num(demand):
                        demand_at_t += demand * ((start[job] <= t) & (t < end[job]))
                    else:
                        demand_at_t += demand[job] * ((start[job] <= t) & (t < end[job]))

                cons += [demand_at_t <= capacity]
                
        elif version == "task":

            # set duration of tasks
            for t in range(num_tasks):
                cons += [start[t] + duration[t] == end[t]]

            for j in range(num_tasks):
                cons += [capacity >= demand[j] + cp.sum([(start[i] <= start[j]) & (start[j] < start[i] + duration[i]) for i in range(num_tasks) if i != j])]

        return cons, []

    def value(self):
        arg_vals = [np.array(argvals(arg)) if is_any_list(arg)
                   else argval(arg) for arg in self.args]

        if any(a is None for a in arg_vals):
            return None

        # start, dur, end are np arrays
        start, dur, end, demand, capacity = arg_vals
        # start and end seperated by duration
        if not (start + dur == end).all():
            return False

        # demand doesn't exceed capacity
        lb, ub = min(start), max(end)
        for t in range(lb, ub+1):
            if capacity < sum(demand * ((start <= t) & (t < end))):
                return False

        return True


# helper function
def is_transition(arg):
    """ test if the argument is a transition, i.e. a 3-elements-tuple specifying a starting state,
    a transition value and an ending node"""
    return len(arg) == 3 and \
        isinstance(arg[0], (int, str)) and is_int(arg[1]) and isinstance(arg[2], (int, str))



# ---------------------------------------------------------------------------- #
#                            Solver-specific Globals                           #
# ---------------------------------------------------------------------------- #


"""
A collection of XCSP3 solver-native global constraints.
"""


# --------------------------------- OR-Tools --------------------------------- #

class OrtNoOverlap2D(DirectConstraint):
    """
    OR-Tools native `NoOverlap2D` global constraint.

    Ensures that all provided rectangles are positioned within a plane whilst not overlapping.
    The rectangles have their sides aligned with the perpendicular x- and y-axis.
    """
    def __init__(self, arguments):
        super().__init__("ortnooverlap2d", arguments)

    def callSolver(self, CPMpy_solver, Native_solver):
        start_x, dur_x, end_x, start_y, dur_y, end_y = CPMpy_solver.solver_vars(self.args[0])
        intervals_x = [Native_solver.NewIntervalVar(s,d,e, f"xinterval_{s}-{d}-{d}") for s,d,e in zip(start_x,dur_x,end_x)]
        intervals_y = [Native_solver.NewIntervalVar(s,d,e, f"yinterval_{s}-{d}-{d}") for s,d,e in zip(start_y,dur_y,end_y)]
        return Native_solver.add_no_overlap_2d(intervals_x, intervals_y)

class OrtSubcircuit(DirectConstraint):
    """
    OR-Tools native `SubCircuit` global constraint.

    A subcircuit is a Hamiltonian path between a subset of the nodes of a graph.
    When a node `i` is not part of the circuit, it should self-loop with an arc `i -> i`.
    """
    def __init__(self, arguments):
        super().__init__("ortsubcircuit", arguments)

    def callSolver(self, CPMpy_solver, Native_solver):
        N = len(self.args[0])
        arcvars = cp.boolvar(shape=(N,N))
        # post channeling constraints from int to bool
        CPMpy_solver.add([b == (self.args[0][i] == j) for (i,j),b in np.ndenumerate(arcvars)])
        # post the global constraint
        #   posting arcs on diagonal (i==j) allows for subcircuits
        ort_arcs = [(i,j, CPMpy_solver.solver_var(b)) for (i,j),b in np.ndenumerate(arcvars)] # Allows for empty subcircuits

        return Native_solver.AddCircuit(ort_arcs)

class OrtSubcircuitWithStart(DirectConstraint):
    """
    OR-Tools native `SubCircuitWithStart` global constraint.

    This global is the same as `SubCircuit`, only can a additional `start_index` be provided 
    which is ensure to be part of the subcircuit.
    """
    def __init__(self, arguments, start_index:int=0):
        super().__init__("ortsubcircuitwithstart", (arguments, start_index))

    def callSolver(self, CPMpy_solver, Native_solver):
        N = len(self.args[0])
        arcvars = cp.boolvar(shape=(N,N))
        # post channeling constraints from int to bool
        CPMpy_solver.add([b == (self.args[0][i] == j) for (i,j),b in np.ndenumerate(arcvars)])
        # post the global constraint
        # posting arcs on diagonal (i==j) allows for subcircuits
        ort_arcs = [(i,j,CPMpy_solver.solver_var(b)) for (i,j),b in np.ndenumerate(arcvars) if not ((i == j) and (i == self.args[1]))] # The start index cannot self loop and thus must be part of the subcircuit.

        return Native_solver.AddCircuit(ort_arcs)
        

# ----------------------------------- Choco ---------------------------------- #

class ChocoSubcircuit(DirectConstraint):
    """
    Choco's native `SubCircuit` global constraint.

    A subcircuit is a Hamiltonian path between a subset of the nodes of a graph.
    When a node `i` is not part of the circuit, it should self-loop with an arc `i -> i`.
    """
    def __init__(self, arguments):
        super().__init__("chocosubcircuit", arguments)

    def callSolver(self, CPMpy_solver, Native_solver):
        # Successor variables
        succ = CPMpy_solver.solver_vars(self.args[0])
        # Add an unused variable for the subcircuit length.
        subcircuit_length = CPMpy_solver.solver_var(cp.intvar(0, len(succ)))
        return Native_solver.sub_circuit(succ, 0, subcircuit_length)

# --------------------------------- Minizinc --------------------------------- #

class MinizincSubcircuit(DirectConstraint):
    """
    Minizinc's native `SubCircuit` global constraint.

    A subcircuit is a Hamiltonian path between a subset of the nodes of a graph.
    When a node `i` is not part of the circuit, it should self-loop with an arc `i -> i`.
    """
    def __init__(self, arguments):
        super().__init__("minizincsubcircuit", arguments)

    def callSolver(self, CPMpy_solver, Native_solver):
        # minizinc is offset 1, which can be problematic here...
        args_str = ["{}+1".format(CPMpy_solver._convert_expression(e)) for e in self.args[0]]
        return "{}([{}])".format("subcircuit", ",".join(args_str))

class MinizincSubcircuitWithStart(DirectConstraint):
    """
    Minizinc's native `SubCircuitWithStart` global constraint.

    This global is the same as `SubCircuit`, only can a additional `start_index` be provided 
    which is ensure to be part of the subcircuit.
    """
    def __init__(self, arguments):
        super().__init__("minizincsubcircuitwithstart", arguments)

    def callSolver(self, CPMpy_solver, Native_solver):
        # minizinc is offset 1, which can be problematic here...
        start_index = self.args[0][-1]
        succ = cpm_array(self.args[0][:-1]) # Successor variables

        CPMpy_solver += (succ[start_index] != start_index)
        args_str = ["{}+1".format(CPMpy_solver._convert_expression(e)) for e in succ]
        return "{}([{}])".format("subcircuit", ",".join(args_str))
        


from cpmpy.tools.explain.mss import *
from cpmpy.transformations.normalize import toplevel_list


def mcs(soft, hard=[], solver="ortools"):
    """
        Compute Minimal Correction Subset of unsatisfiable model.
        Removing these contraints will result in a satisfiable model.
        Computes a subset of constraints which minimizes the total number of constraints to be removed

        :param: soft: list of soft constraints that may be part of the minimal correction subset
        :param: hard: list of hard constraints, will be added to the model before solving
        :param: solver: the SAT-solver to use, must support optimization
    """
    return mcs_opt(soft, hard, 1, solver)


def mcs_opt(soft, hard, weights=1, solver="ortools"):
    """
        Compute Minimal Correction Subset of unsatisfiable model.
        Constraints can be weighted using the `weights` parameter.
        Computes a subset of constraints which minimizes the sum of all weights of constraints.

        :param: soft: list of soft constraints that may be part of the minimal correction subset
        :param: hard: list of hard constraints, will be added to the model before solving
        :param: weights: weight of each constraint, default is 1
        :param: solver: the SAT-solver to use, must support optimization
    """
    soft2 = toplevel_list(soft, merge_and=False)
    mymss = mss_opt(soft2, hard, weights, solver=solver)
    return list(set(soft2) - set(mymss))


def mcs_grow(soft, hard, solver="ortools"):
    """
        Computes correction subset without requirement of optimization support
        Relies on assumptions so incremental solvers are adviced.
        Can be faster in some cases compared to optimal correction subset

        :param: soft: list of soft constraints that may be part of the minimal correction subset
        :param: hard: list of hard constraints, will be added to the model before solving
        :param: solver: the SAT-solver to use, must support assumptions
    """
    soft2 = toplevel_list(soft, merge_and=False)
    mymss = mss_grow(soft2, hard, solver)
    return list(set(soft2) - set(mymss))


def mcs_grow_naive(soft, hard, solver="ortools"):
    """
        Compute Minimal Correction Subset of unsatsifiable model.
        Computes a subset-minimal set of constraints by greedily removing contraints.
        Can be used when solver does not support assumptions
        No guarantees on optimality, but can be faster in some cases

        :param: soft: list of soft constraints that may be part of the minimal correction subset
        :param: hard: list of hard constraints, will be added to the model before solving
        :param: solver: the SAT-solver to use, ideally incremental such as "gurobi", "exact"
    """
    soft2 = toplevel_list(soft, merge_and=False)
    mymss = mss_grow_naive(soft2, hard, solver)
    return list(set(soft2) - set(mymss))



"""
    Transforms partial functions into total functions.
"""

from copy import copy

from ..expressions.variables import _NumVarImpl, boolvar, intvar, NDVarArray, cpm_array
from ..expressions.core import Expression, Operator, BoolVal
from ..expressions.utils import get_bounds, is_num
from ..expressions.globalfunctions import GlobalFunction, Element
from ..expressions.globalconstraints import DirectConstraint
from ..expressions.python_builtins import all as cpm_all

def no_partial_functions(lst_of_expr, _toplevel=None, _nbc=None, safen_toplevel=frozenset()):
    """
        A partial function is a function whose output is not defined for all possible inputs.

        In CPMpy, this is the case for the following 3 numeric functions:

        - (Integer) division ``x // y``: undefined when y=0
        - Modulo ``x mod y``: undefined when y=0
        - Element ``Arr[idx]``: undefined when idx is not in the range of Arr

        A toplevel constraint must always be true, so constraint solvers simply propagate the 'unsafe'
        value(s) away. However, CPMpy allows arbitrary nesting and reification of constraints, so an
        expression like ``b <-> (Arr[idx] == 5)`` is allowed, even when variable `idx` goes outside the bounds of `Arr`.
        Should `idx` be restricted to be in-bounds? and what value should 'b' take if it is out-of-bounds?

        This transformation will transform a partial function (e.g. Arr[idx]) into a total function
        following the **relational semantics** as discussed in:
            Frisch, Alan M., and Peter J. Stuckey. "The proper treatment of undefinedness in
            constraint languages." International Conference on Principles and Practice of Constraint
            Programming. Berlin, Heidelberg: Springer Berlin Heidelberg, 2009.

        Under the relational semantic, an 'undefined' output for a (numerical) expression should
        propagate to `False` in the nearest Boolean parent expression. In the above example: `idx` should
        be allowed to take a value outside the bounds of `Arr`, and `b` should be False in that case.

        To enable this, we want to rewrite an expression like ``b <-> (partial == 5)`` to something like
        ``b <-> (is_defined & (total == 5))``. The key idea is to create a copy of the potentially unsafe
        argument, that can only take 'safe' values. Using this new argument in the original expression results
        in a total function. We now have the original argument variable, which is decoupled from the expression,
        and a new 'safe' argument variable which is used in the expression. The `is_defined` flag serves two
        purposes: it represents whether the original argument has a safe value; and if it is true the new
        argument must equal the original argument so the two are coupled again. If `is_defined` is false, the new
        argument remains decoupled (can take any value, as will the function's output).

        .. warning::
            Under the relational semantics, ``b <-> ~(partial==5)`` and ``b <-> (partial!=5)`` mean
            different things! The second is ``b <-> (is_defined & (total!=5))`` the first is
            ``b <-> (~is_defined | (total!=5))``.


        A clever observation of the implementation below is that for the above 3 expressions, the 'safe'
        domain of a potentially unsafe argument (y or idx) is either one 'trimmed' continuous domain
        (for idx and in case y = [0..n] or [-n..0]), or two 'trimmed' continuous domains (for y=[-n..m]).
        Furthermore, the reformulation for these two cases can be done generically, without needing
        to know the specifics of the partial function being made total.


        :param list_of_expr: list of CPMpy expressions
        :param _toplevel: list of new expressions to put toplevel (used internally)
        :param _nbc: list of new expressions to put in nearest Boolean context (used internally)
        :param safen toplevel: list of expression types that need to be safened, even when toplevel. Used when
                                 a solver does not support unsafe values in it's API (e.g., OR-Tools for `div`).
    """

    if _toplevel is None:
        assert _nbc is None, f"_nbc is an internal argument, should not be filled by caller but got {_nbc}"
        toplevel_call = True
        _toplevel = []
        _nbc = _toplevel # at the toplevel of the contraint model, the neirest Boolean context is just toplevel
    else:
        toplevel_call = False
        assert isinstance(_toplevel, list), f"_toplevel argument must be of type list but got {type(_toplevel)}"
        assert isinstance(_nbc, list), f"_nbc argument must be of type list but got {type(_nbc)}"

    new_lst = []
    for cpm_expr in lst_of_expr:

        if is_num(cpm_expr) or isinstance(cpm_expr, _NumVarImpl):
            new_lst.append(cpm_expr)

        elif isinstance(cpm_expr, (list,tuple)):
            new_lst.append(no_partial_functions(cpm_expr, _toplevel, _nbc, safen_toplevel))

        elif isinstance(cpm_expr, NDVarArray):
            if cpm_expr.has_subexpr():
                # efficiency: flatten into single iterator, then reshape to n-dimensional again
                new_cpm_expr = cpm_array(no_partial_functions(cpm_expr.flat, _toplevel, _nbc, safen_toplevel)).reshape(cpm_expr.shape)
                new_lst.append(new_cpm_expr)
            else:
                new_lst.append(cpm_expr)

        elif isinstance(cpm_expr, DirectConstraint):  # do not recurse into args
            new_lst.append(cpm_expr)

        else:
            assert isinstance(cpm_expr, Expression), f"each `cpm_expr` should be an Expression at this point, not {type(cpm_expr)}"

            args = cpm_expr.args

            if cpm_expr.is_bool() and toplevel_call is False:  # a Boolean context, create a new _nbc
                # if we are currently toplevel, no need to create a new nbc, will append to toplevel anyway
                _nbc = []

            # recurse over the arguments of the expression
            if cpm_expr.has_subexpr():
                new_args = no_partial_functions(args, _toplevel, _nbc, safen_toplevel=safen_toplevel)
                if any((a1 is not a2) for a1,a2 in zip(new_args,args)):  # efficiency (hopefully): only copy if an arg changed
                    cpm_expr = copy(cpm_expr)
                    cpm_expr.update_args(new_args)
                    args = new_args

            if cpm_expr.is_bool() and len(_nbc) != 0 and toplevel_call is False:
                # a nested Boolean context, conjoin my Boolean expression with _nbc
                # in `b <-> (Arr[idx] == 5)`, this would trigger for cpm_expr (Arr[idx] == 5)
                cpm_expr = cpm_all(_nbc) & cpm_expr

            elif isinstance(cpm_expr, GlobalFunction) and cpm_expr.name == "element":
                if _nbc is _toplevel and cpm_expr.name not in safen_toplevel: # no need to safen
                    new_lst.append(cpm_expr)
                    continue

                arr, idx = args
                lb, ub = get_bounds(idx)

                if lb < 0 or ub >= len(arr): # index can be out of bounds
                    guard, output_expr, extra_cons = _safen_range(cpm_expr, safe_range=(0, len(arr)-1), idx_to_safen=1)

                    _nbc.append(guard)  # guard must be added to nearest Boolean context
                    _toplevel += extra_cons  # any additional constraint that must be true
                    cpm_expr = output_expr  # replace partial function by this (total) new output expression

            elif cpm_expr.name == "div" or cpm_expr.name == "mod":
                if _nbc is _toplevel and cpm_expr.name not in safen_toplevel: # no need to safen
                    new_lst.append(cpm_expr)
                    continue

                idx_to_safen = 1
                lb, ub = get_bounds(args[idx_to_safen])

                if lb <= 0 <= ub:
                    if lb == ub == 0:
                        # (unlikely) edge case, neirest Boolean context should propagate to False.
                        # introduce dummy numerical integer expression
                        guard, output_expr, extra_cons = BoolVal(False), intvar(0,1), []
                    elif lb == 0:
                        guard, output_expr, extra_cons = _safen_range(cpm_expr, safe_range=(1, ub), idx_to_safen=idx_to_safen)
                    elif ub == 0:
                        guard, output_expr, extra_cons = _safen_range(cpm_expr, safe_range=(lb, -1), idx_to_safen=idx_to_safen)
                    else:  # proper hole
                        guard, output_expr, extra_cons = _safen_hole(cpm_expr, exclude=0, idx_to_safen=idx_to_safen)

                    _nbc.append(guard)  # guard must be added to nearest Boolean context
                    _toplevel += extra_cons  # any additional constraint that must be true
                    cpm_expr = output_expr  # replace partial function by this (total) new output expression

            # add the (potentially new) cpm_expr to the list
            new_lst.append(cpm_expr)

    if toplevel_call is True:
        assert _nbc is _toplevel # should point to the same list here
        return new_lst + _toplevel
    else:
        return new_lst



def _safen_range(partial_expr, safe_range, idx_to_safen):
    """
        Replace partial function `cpm_expr` that has potentially unsafe argument at `idx_to_safen`,
        by a total function using a safe argument with domain `safe_range`. Also returns
        a Boolean flag indicating whether the original argument's value is in the safe range
        (for use in the nearest Boolean context), and a list of toplevel constraints that help
        define the new total function.

        An example is `Element` where the index should be within the array's range.

        :param partial_expr: The partial function expression to make total
        :param safe_range: The range of safe argument values
        :param idx_to_safen: The index of the potentially unsafe argument in the expression

    """
    safe_lb, safe_ub = safe_range

    orig_arg = partial_expr.args[idx_to_safen]
    new_arg = intvar(safe_lb, safe_ub)  # values for which the partial function is defined

    total_expr = copy(partial_expr)  # the new total function, with the new arg
    total_expr.update_args([new_arg if i == idx_to_safen else a for i,a in enumerate(partial_expr.args)])

    is_defined = boolvar()
    toplevel = [is_defined == ((safe_lb <= orig_arg) & (orig_arg <= safe_ub)),
                is_defined.implies(new_arg == orig_arg),
                # Extra: avoid additional solutions when new var is unconstrained
                # (should always work in reified context because total_expr is newly defined?)
                (~is_defined).implies(new_arg == safe_lb)]

    return is_defined, total_expr, toplevel


def _safen_hole(cpm_expr, exclude, idx_to_safen):
    """
        Safen expression where a single value of an argument can cause undefinedness.
        Examples include `div` where 0 has to be removed from the denominator

        Constructs an expression for each interval of safe values, and
        introduces a new `output_var` variable

        :param cpm_expr: The numerical expression to safen
        :param exclude: The domain value to exclude
        :param idx_to_safen: The index of unsafe argument in the expression
    """
    orig_arg = cpm_expr.args[idx_to_safen]
    orig_lb, orig_ub = get_bounds(orig_arg)

    # expr when arg in [orig_lb..exclude-1]
    new_arg_lower = intvar(orig_lb, exclude-1)
    total_expr_lower = copy(cpm_expr)
    total_expr_lower.update_args([new_arg_lower if i == idx_to_safen else a for i,a in enumerate(cpm_expr.args)])
    # expr when arg in [exclude+1..orig_ub]
    new_arg_upper = intvar(exclude+1, orig_ub)
    total_expr_upper = copy(cpm_expr)
    total_expr_upper.update_args([new_arg_upper if i == idx_to_safen else a for i, a in enumerate(cpm_expr.args)])

    is_defined = boolvar()
    is_defined_lower = boolvar()
    is_defined_upper = boolvar()
    output_var = intvar(*get_bounds(cpm_expr))

    toplevel = [
        is_defined_lower == (orig_arg < exclude),
        is_defined_upper == (orig_arg > exclude),
        is_defined == (is_defined_lower | is_defined_upper),

        is_defined_lower.implies(orig_arg == new_arg_lower),
        is_defined_lower.implies(output_var == total_expr_lower),
        is_defined_upper.implies(orig_arg == new_arg_upper),
        is_defined_upper.implies(output_var == total_expr_upper),

        # Extra: avoid additional solutions when new vars are unconstrained
        # (should always work in reified context because total_expr's are newly defined?)
        (~is_defined_lower).implies(new_arg_lower == exclude-1),
        (~is_defined_upper).implies(new_arg_upper == exclude+1),
        (~is_defined).implies(output_var == output_var.lb)
    ]

    return is_defined, output_var, toplevel





#!/usr/bin/env python
#-*- coding:utf-8 -*-
##
## pysat.py
##
"""
    Interface to PySAT's API

    PySAT is a Python (2.7, 3.4+) toolkit, which aims at providing a simple and unified
    interface to a number of state-of-art Boolean satisfiability (SAT) solvers as well as
    to a variety of cardinality and pseudo-Boolean encodings.
    (see https://pysathq.github.io/)

    .. warning::
        This solver can only be used if the model only uses Boolean variables.
        It does not support optimization.
    
    Always use :func:`cp.SolverLookup.get("pysat") <cpmpy.solvers.utils.SolverLookup.get>` to instantiate the solver object.

    ============
    Installation
    ============

    Requires that the 'python-sat' package is installed:

    .. code-block:: console

        $ pip install python-sat

    If you want to also solve pseudo-Boolean constraints, you should also install its optional dependency 'pypblib', as follows:

    .. code-block:: console

        $ pip install pypblib

    See detailed installation instructions at:
    https://pysathq.github.io/installation

    The rest of this documentation is for advanced users.

    ===============
    List of classes
    ===============

    .. autosummary::
        :nosignatures:

        CPM_pysat

    ==============
    Module details
    ==============
"""
from threading import Timer
from typing import Optional
import warnings
import pkg_resources

from .solver_interface import SolverInterface, SolverStatus, ExitStatus
from ..exceptions import NotSupportedError
from ..expressions.core import Comparison, Operator, BoolVal
from ..expressions.variables import _BoolVarImpl, _IntVarImpl, NegBoolView
from ..expressions.globalconstraints import DirectConstraint
from ..transformations.linearize import only_positive_coefficients
from ..expressions.utils import flatlist
from ..transformations.decompose_global import decompose_in_tree
from ..transformations.get_variables import get_variables
from ..transformations.flatten_model import flatten_constraint
from ..transformations.linearize import linearize_constraint
from ..transformations.normalize import toplevel_list, simplify_boolean
from ..transformations.reification import only_implies, only_bv_reifies
from ..transformations.int2bool import int2bool, _encode_int_var, _decide_encoding


class CPM_pysat(SolverInterface):
    """
    Interface to PySAT's API.

    Creates the following attributes (see parent constructor for more):

    - ``pysat_vpool``: a pysat.formula.IDPool for the variable mapping
    - ``pysat_solver``: a pysat.solver.Solver() (default: glucose4)
    - ``ivarmap``: a mapping from integer variables to their encoding for `int2bool`
    - ``encoding``: the encoding used for `int2bool`, choose from ("auto", "direct", "order", or "binary"). Set to "auto" but can be changed in the solver object.

    The :class:`~cpmpy.expressions.globalconstraints.DirectConstraint`, when used, calls a function on the ``pysat_solver`` object.

    Documentation of the solver's own Python API:
    https://pysathq.github.io/docs/html/api/solvers.html

    .. note::
        CPMpy uses 'model' to refer to a constraint specification,
        the PySAT docs use 'model' to refer to a solution.

    """

    @staticmethod
    def supported():
        # try to import the package
        try:
            import pysat
            # there is actually a non-related 'pysat' package
            # while we need the 'python-sat' package, some more checks:
            from pysat.formula import IDPool
            from pysat.solvers import Solver

            from pysat import card
            CPM_pysat._card = card  # native

            # try to import pypblib and avoid ever re-import by setting `_pb`
            if not hasattr(CPM_pysat, ("_pb")):
                try:
                    from pysat import pb  # require pypblib
                    """The `pysat.pb` module if its dependency `pypblib` installed, `None` if we have not checked it yet, or `False` if we checked and it is *not* installed"""
                    CPM_pysat._pb = pb
                except (ModuleNotFoundError, NameError):  # pysat returns the wrong error type (latter i/o former)
                    CPM_pysat._pb = None  # not installed, avoid reimporting

            return True
        except ModuleNotFoundError:
            return False
        except Exception as e:
            raise e

    @staticmethod
    def solvernames(**kwargs):
        """
            Returns solvers supported by PySAT on your system
        """
        if CPM_pysat.supported():
            from pysat.solvers import SolverNames
            names = []
            for name, attr in vars(SolverNames).items():
                # issue with cryptosat, so we don't include it in our https://github.com/msoos/cryptominisat/issues/765
                if not name.startswith('__') and isinstance(attr, tuple) and name != 'cryptosat':
                    if name not in attr:
                        name = attr[-1]
                    names.append(name)  
            return names
        else:
            warnings.warn("PySAT is not installed or not supported on this system.")
            return []
        
    @staticmethod
    def solverversion(subsolver:str) -> Optional[str]:
        """
        Returns the version of the requested subsolver.

        Arguments:
            subsolver (str): name of the subsolver

        Returns:
            Version number of the subsolver if installed, else None 
    
        Pysat currently does not provide accessible subsolver version numbers.
        """
        # Could try to extract them from solver name, but even then the minor revision numbers are missing
        return None
    
    @staticmethod
    def version() -> Optional[str]:
        """
        Returns the installed version of the solver's Python API.
        """
        try:
            return pkg_resources.get_distribution('python-sat').version
        except pkg_resources.DistributionNotFound:
            return None

    def __init__(self, cpm_model=None, subsolver=None):
        """
        Constructor of the native solver object

        Requires a CPMpy model as input, and will create the corresponding
        PySAT clauses and solver object

        Only supports satisfaction problems (no objective)

        Arguments:
            cpm_model (Model(), a CPMpy Model(), optional):
            subsolver (str, name of the pysat solver, e.g. glucose4):  see .solvernames() to get the list of available solver(names)
        """
        if not self.supported():
            raise ImportError("PySAT is not installed. The recommended way to install PySAT is with `pip install cpmpy[pysat]`, or `pip install python-sat` if you do not require `pblib` to encode (weighted) sums.")
        if cpm_model and cpm_model.objective_ is not None:
            raise NotSupportedError("CPM_pysat: only satisfaction, does not support an objective function")

        from pysat.formula import IDPool
        from pysat.solvers import Solver

        # determine subsolver
        if subsolver is None or subsolver == 'pysat':
            # default solver
            subsolver = "glucose4" # something recent...
        elif subsolver.startswith('pysat:'):
            subsolver = subsolver[6:] # strip 'pysat:'

        # initialise the native solver object
        self.pysat_vpool = IDPool()
        self.pysat_solver = Solver(use_timer=True, name=subsolver)
        self.ivarmap = dict()  # for the integer to boolean encoders
        self.encoding = "auto"

        # initialise everything else and post the constraints/objective
        super().__init__(name="pysat:"+subsolver, cpm_model=cpm_model)

    @property
    def native_model(self):
        """
            Returns the solver's underlying native model (for direct solver access).
        """
        return self.pysat_solver


    def solve(self, time_limit=None, assumptions=None):
        """
            Call the PySAT solver

            Arguments:
                time_limit (float, optional):   Maximum solve time in seconds. Auto-interrups in case the
                                                runtime exceeds given time_limit.
                                                
                                                .. warning::
                                                    Warning: the time_limit is not very accurate at subsecond level
                assumptions: list of CPMpy Boolean variables that are assumed to be true.
                            For use with :func:`s.get_core() <get_core()>`: if the model is UNSAT, get_core() returns a small subset of assumption variables that are unsat together.
                            Note: the PySAT interface is statefull, so you can incrementally call solve() with assumptions and it will reuse learned clauses
        """

        # ensure all vars are known to solver
        self.solver_vars(list(self.user_vars))

        # the user vars are only the Booleans (e.g. to ensure solveAll behaves consistently)
        user_vars = set()
        for x in self.user_vars:
            if isinstance(x, _BoolVarImpl):
                user_vars.add(x)
            else:
                # extends set with encoding variables of `x`
                user_vars.update(self.ivarmap[x.name].vars())

        self.user_vars = user_vars

        if assumptions is None:
            pysat_assum_vars = [] # default if no assumptions
        else:
            pysat_assum_vars = self.solver_vars(assumptions)
            self.assumption_vars = assumptions

        # set time limit
        if time_limit is not None:
            if time_limit <= 0:
                raise ValueError("Time limit must be positive")
            
            t = Timer(time_limit, lambda s: s.interrupt(), [self.pysat_solver])
            t.start()
            my_status = self.pysat_solver.solve_limited(assumptions=pysat_assum_vars, expect_interrupt=True)
            # ensure timer is stopped if early stopping
            t.cancel()
            ## this part cannot be added to timer otherwhise it "interrups" the timeout timer too soon
            self.pysat_solver.clear_interrupt()
        else:
            my_status = self.pysat_solver.solve(assumptions=pysat_assum_vars)

        # new status, translate runtime
        self.cpm_status = SolverStatus(self.name)
        self.cpm_status.runtime = self.pysat_solver.time()

        # translate exit status
        if my_status is True:
            # COP
            if self.has_objective():
                self.cpm_status.exitstatus = ExitStatus.OPTIMAL
            # CSP
            else:
                self.cpm_status.exitstatus = ExitStatus.FEASIBLE
        elif my_status is False:
            self.cpm_status.exitstatus = ExitStatus.UNSATISFIABLE
        elif my_status is None:
            # can happen when timeout is reached...
            self.cpm_status.exitstatus = ExitStatus.UNKNOWN
        else:  # another?
            raise NotImplementedError(my_status)  # a new status type was introduced, please report on github

        # True/False depending on self.cpm_status
        has_sol = self._solve_return(self.cpm_status)

        # translate solution values (of user specified variables only)
        if has_sol:
            sol = frozenset(self.pysat_solver.get_model())  # to speed up lookup
            # fill in variable values
            for cpm_var in self.user_vars:
                if isinstance(cpm_var, _BoolVarImpl):
                    lit = self.solver_var(cpm_var)
                    if lit in sol:
                        cpm_var._value = True
                    else:  # -lit in sol (=False) or not specified (=False)
                        cpm_var._value = False
                elif isinstance(cpm_var, _IntVarImpl):
                    raise TypeError("user_vars should only contain Booleans")
                else:
                    raise NotImplementedError(f"CPM_pysat: variable {cpm_var} not supported")

            # Now assign the user integer variables using their encodings
            # `ivarmap` also contains auxiliary variable, but they will be assigned 'None' as their encoding variables are assigned `None`
            for enc in self.ivarmap.values():
                enc._x._value = enc.decode()

        else: # clear values of variables
            for cpm_var in self.user_vars:
                cpm_var._value = None


        return has_sol


    def solver_var(self, cpm_var):
        """
            Creates solver variable for cpmpy variable
            or returns from cache if previously created.

            Transforms cpm_var into CNF literal using ``self.pysat_vpool``
            (positive or negative integer).

            So vpool is the varmap (we don't use _varmap here).
        """

        # special case, negative-bool-view
        # work directly on var inside the view
        if isinstance(cpm_var, BoolVal):
            return cpm_var
        elif isinstance(cpm_var, NegBoolView):
            # just a view, get actual var identifier, return -id
            return -self.pysat_vpool.id(cpm_var._bv.name)
        elif isinstance(cpm_var, _BoolVarImpl):
            return self.pysat_vpool.id(cpm_var.name)
        elif isinstance(cpm_var, _IntVarImpl):  # intvar
            if cpm_var.name not in self.ivarmap:
                enc, cons = _encode_int_var(self.ivarmap, cpm_var, _decide_encoding(cpm_var, None, encoding=self.encoding))
                self += cons
            else:
                enc = self.ivarmap[cpm_var.name]
            return self.solver_vars(enc.vars())
        else:
            raise NotImplementedError(f"CPM_pysat: variable {cpm_var} not supported")

    def transform(self, cpm_expr):
        """
            Transform arbitrary CPMpy expressions to constraints the solver supports

            Implemented through chaining multiple solver-independent **transformation functions** from
            the `cpmpy/transformations/` directory.

            See the :ref:`Adding a new solver` docs on readthedocs for more information.


            In the case of PySAT, the supported constraints are over Boolean variables:

            - Boolean clauses
            - Cardinality constraint (`sum`)
            - Pseudo-Boolean constraints (`wsum`)

            :param cpm_expr: CPMpy expression, or list thereof
            :type cpm_expr: Expression or list of Expression

            :return: list of Expression
        """
        cpm_cons = toplevel_list(cpm_expr)
        cpm_cons = decompose_in_tree(cpm_cons, supported=frozenset({"alldifferent"}), csemap=self._csemap)
        cpm_cons = simplify_boolean(cpm_cons)
        cpm_cons = flatten_constraint(cpm_cons, csemap=self._csemap)  # flat normal form
        cpm_cons = only_bv_reifies(cpm_cons, csemap=self._csemap)
        cpm_cons = only_implies(cpm_cons, csemap=self._csemap)
        cpm_cons = linearize_constraint(cpm_cons, supported=frozenset({"sum","wsum", "and", "or"}), csemap=self._csemap)  # the core of the MIP-linearization
        cpm_cons = int2bool(cpm_cons, self.ivarmap, encoding=self.encoding)
        cpm_cons = only_positive_coefficients(cpm_cons)
        return cpm_cons

    def add(self, cpm_expr_orig):
        """
            Eagerly add a constraint to the underlying solver.

            Any CPMpy expression given is immediately transformed (through `transform()`)
            and then posted to the solver in this function.

            This can raise 'NotImplementedError' for any constraint not supported after transformation

            The variables used in expressions given to add are stored as 'user variables'. Those are the only ones
            the user knows and cares about (and will be populated with a value after solve). All other variables
            are auxiliary variables created by transformations.

            What 'supported' means depends on the solver capabilities, and in effect on what transformations
            are applied in `transform()`.
        """
        # add new user vars to the set
        get_variables(cpm_expr_orig, collect=self.user_vars)

        # transform and post the constraints
        for cpm_expr in self.transform(cpm_expr_orig):
            self._post_constraint(cpm_expr)

        return self

    def _post_constraint(self, cpm_expr):
        """ Add expression to solver _without_ transforming."""
        if cpm_expr.name == 'or':
            self.pysat_solver.add_clause(self.solver_vars(cpm_expr.args))

        elif cpm_expr.name == '->':  # BV -> BE only thanks to only_bv_reifies
            a0,a1 = cpm_expr.args

            if isinstance(a1, _BoolVarImpl):
                # BoolVar() -> BoolVar()
                args = [~a0, a1]
                self.pysat_solver.add_clause(self.solver_vars(args))
            elif isinstance(a1, Operator) and a1.name == 'or':
                # BoolVar() -> or(...)
                args = [~a0]+a1.args
                self.pysat_solver.add_clause(self.solver_vars(args))
            elif isinstance(a1, Comparison) and a1.args[0].name == "sum":  # implied sum comparison (a0->sum(bvs)<>val)
                # implied sum comparison (a0->sum(bvs)<>val)
                # convert sum to cnf
                cnf = self._pysat_cardinality(a1, reified=True)
                # implication of conjunction is conjunction of individual implications
                antecedent = [self.solver_var(~a0)]
                cnf = [antecedent+c for c in cnf]
                self.pysat_solver.append_formula(cnf)
            elif isinstance(a1, Comparison) and a1.args[0].name == "wsum":  # implied pseudo-boolean comparison (a0->wsum(ws,bvs)<>val)
                # implied sum comparison (a0->wsum([w,bvs])<>val or a0->(w*bv<>val))
                cnf = self._pysat_pseudoboolean(a1)
                # implication of conjunction is conjunction of individual implications
                antecedent = [self.solver_var(~a0)]
                self.pysat_solver.append_formula([antecedent+c for c in cnf])
            else:
                raise NotSupportedError(f"Implication: {cpm_expr} not supported by CPM_pysat")

        elif isinstance(cpm_expr, Comparison): # root-level comparisons have been linearized
            if isinstance(cpm_expr.args[0], Operator) and cpm_expr.args[0].name == "sum":
                self.pysat_solver.append_formula(self._pysat_cardinality(cpm_expr))
            elif isinstance(cpm_expr.args[0], Operator) and cpm_expr.args[0].name == "wsum":
                self.pysat_solver.append_formula(self._pysat_pseudoboolean(cpm_expr))
            else:
                raise NotSupportedError(f"Implication: {cpm_expr} not supported by CPM_pysat")

        elif isinstance(cpm_expr, Comparison):
            # comparisons between Booleans will have been transformed out
            # check if comparison of cardinality/pseudo-boolean constraint
            if isinstance(cpm_expr.args[0], Operator):
                if cpm_expr.args[0].name == "sum":
                    # convert to clauses and post
                    clauses = self._pysat_cardinality(cpm_expr)
                    self.pysat_solver.append_formula(clauses)
                elif cpm_expr.args[0].name == "wsum":
                    # convert to clauses and post
                    clauses = self._pysat_pseudoboolean(cpm_expr)
                    self.pysat_solver.append_formula(clauses)
                else:
                    raise NotImplementedError(f"Operator constraint {cpm_expr} not supported by CPM_pysat")
            else:
                raise NotImplementedError(f"Non-operator constraint {cpm_expr} not supported by CPM_pysat")

        elif isinstance(cpm_expr, BoolVal):
            # base case: Boolean value
            if cpm_expr.args[0] is False:
                self.pysat_solver.add_clause([])

        elif isinstance(cpm_expr, _BoolVarImpl):
            # base case, just var or ~var
            self.pysat_solver.add_clause([self.solver_var(cpm_expr)])

        # a direct constraint, pass to solver
        elif isinstance(cpm_expr, DirectConstraint):
            cpm_expr.callSolver(self, self.pysat_solver)

        else:
            raise NotImplementedError(f"CPM_pysat: Non supported constraint {cpm_expr}")


    __add__ = add  # avoid redirect in superclass

    def solution_hint(self, cpm_vars, vals):
        """
        PySAT supports warmstarting the solver with a feasible solution

        In PySAT, this is called setting the 'phases' or the 'polarities' of literals

        :param cpm_vars: list of CPMpy variables
        :param vals: list of (corresponding) values for the variables
        """

        cpm_vars = flatlist(cpm_vars)
        vals = flatlist(vals)
        assert (len(cpm_vars) == len(vals)), "Variables and values must have the same size for hinting"

        literals = []
        for (cpm_var, val) in zip(cpm_vars, vals):
            lit = self.solver_var(cpm_var)
            if val:
                # true, so positive literal
                literals.append(lit)
            else:
                # false, so negative literal
                literals.append(-lit)
        self.pysat_solver.set_phases(literals)


    def get_core(self):
        """
            For use with :func:`s.solve(assumptions=[...]) <solve()>`. Only meaningful if the solver returned UNSAT. In that case, get_core() returns a small subset of assumption variables that are unsat together.

            CPMpy will return only those assumptions which are False (in the UNSAT core)

            Note that there is no guarantee that the core is minimal.
            More advanced Minimal Unsatisfiable Subset are available in the 'examples' folder on GitHub

        """
        assert hasattr(self, 'assumption_vars'), "get_core(): requires a list of assumption variables, e.g. s.solve(assumptions=[...])"
        assert (self.cpm_status.exitstatus == ExitStatus.UNSATISFIABLE), "get_core(): solver must return UNSAT"

        assum_idx = frozenset(self.pysat_solver.get_core()) # to speed up lookup

        return [v for v in self.assumption_vars if self.solver_var(v) in assum_idx]


    def _pysat_cardinality(self, cpm_expr, reified=False):
        """ Convert CPMpy comparison of `sum` (over Boolean variables) into PySAT list of clauses """

        # unpack and transform to PySAT argument
        lhs, rhs = cpm_expr.args
        if lhs.name != "sum":
            raise NotSupportedError(
                f"PySAT: Expect {cpm_expr} to be a 'sum'"
            )

        lits = self.solver_vars(lhs.args)
        pysat_args = { "lits": lits, "bound": rhs, "vpool": self.pysat_vpool }

        # Some subsolvers (e.g. MiniCard) support native root context cardinality constraints
        if not reified and self.pysat_solver.supports_atmost():
            pysat_args["encoding"] = self._card.EncType.native

        if cpm_expr.name == "<=":
            return self._card.CardEnc.atmost(**pysat_args)
        elif cpm_expr.name == ">=":
            return self._card.CardEnc.atleast(**pysat_args)
        elif cpm_expr.name == "==":
            return self._card.CardEnc.equals(**pysat_args)
        else:
            raise ValueError(f"PySAT: Expected Comparison to be either <=, ==, or >=, but was {cpm_expr.name}")

    def _pysat_pseudoboolean(self, cpm_expr):
        """Convert CPMpy comparison of `wsum` (over Boolean variables) into PySAT list of clauses."""
        if self._pb is None:
            raise ImportError("The model contains a PB constraint, for which PySAT needs an additional dependency (PBLib). To install it, run `pip install pypblib`.")

        if cpm_expr.args[0].name != "wsum":
            raise NotSupportedError(
                f"PySAT: Expect {cpm_expr} to be a 'wsum'"
            )

        # unpack and transform to PySAT arguments
        lhs, rhs = cpm_expr.args
        lits = self.solver_vars(lhs.args[1])
        pysat_args = {"weights": lhs.args[0], "lits": lits, "bound": rhs, "vpool":self.pysat_vpool }


        if cpm_expr.name == "<=":
            return self._pb.PBEnc.atmost(**pysat_args).clauses
        elif cpm_expr.name == ">=":
            return self._pb.PBEnc.atleast(**pysat_args).clauses
        elif cpm_expr.name == "==":
            return self._pb.PBEnc.equals(**pysat_args).clauses
        else:
            raise ValueError(f"PySAT: Expected Comparison to be either <=, ==, or >=, but was {cpm_expr.name}")



"""
    Transformations are used by solvers to convert (high-level) CPMpy expressions
    into the low-level constraints they support.

    Typical users never need to use these functions directly.

    **CPMpy’s transformations** selectively rewrite only those constraint expressions that a solver does not support. While solvers can use any transformation they need, lower-level solvers largely reuse those of higher-level ones, creating a waterfall pattern:

    .. image:: ../waterfall.png
        :width: 480
        :alt: Waterfall from model to solvers

    The rest of this documentation is for solver developers and other advanced users.
    Make sure you read `Adding a new solver <../adding_solver.html>`_ first.
     
    A transformation can not modify expressions in-place but in that case
    should create and return new expression objects (copy-on-write). In this way, the
    expressions prior to the transformation remain intact, and could be
    used for other purposes too.

    ==================
    List of submodules
    ==================

    Input and output to transformations are always CPMpy expressions, so transformations can
    be chained and called multiple times, as needed. While there is no fixed ordering,
    the following ordering corresponds to the waterfall:

    .. autosummary::
        :nosignatures:

        get_variables
        normalize
        safening
        decompose_global
        negation
        flatten_model
        reification
        comparison
        linearize
        to_cnf
"""



#!/usr/bin/env python
#-*- coding:utf-8 -*-
##
## ortools.py
##
"""
    Interface to OR-Tools' CP-SAT Python API. 

    Google OR-Tools is open source software for combinatorial optimization, which seeks
    to find the best solution to a problem out of a very large set of possible solutions.
    The OR-Tools CP-SAT solver is an award-winning constraint programming solver
    that uses SAT (satisfiability) methods and lazy-clause generation 
    (see https://developers.google.com/optimization).

    Always use :func:`cp.SolverLookup.get("ortools") <cpmpy.solvers.utils.SolverLookup.get>` to instantiate the solver object.

    ============
    Installation
    ============
    
    The 'ortools' python package is bundled by default with CPMpy.
    It can also be installed separately through `pip`:

    .. code-block:: console
    
        $ pip install ortools

    Detailed installation instructions available at:
    https://developers.google.com/optimization/install

    The rest of this documentation is for advanced users.

    ===============
    List of classes
    ===============

    .. autosummary::
        :nosignatures:

        CPM_ortools

    ==============
    Module details
    ==============
"""
import sys
from typing import Optional  # for stdout checking
import numpy as np
import pkg_resources

from .solver_interface import SolverInterface, SolverStatus, ExitStatus
from ..exceptions import NotSupportedError
from ..expressions.core import Expression, Comparison, Operator, BoolVal
from ..expressions.globalconstraints import DirectConstraint
from ..expressions.variables import _NumVarImpl, _IntVarImpl, _BoolVarImpl, NegBoolView, boolvar, intvar
from ..expressions.globalconstraints import GlobalConstraint
from ..expressions.utils import is_num, is_int, eval_comparison, flatlist, argval, argvals, get_bounds, is_true_cst, \
    is_false_cst
from ..transformations.decompose_global import decompose_in_tree
from ..transformations.get_variables import get_variables
from ..transformations.flatten_model import flatten_constraint, flatten_objective, get_or_make_var
from ..transformations.normalize import toplevel_list
from ..transformations.reification import only_implies, reify_rewrite, only_bv_reifies
from ..transformations.comparison import only_numexpr_equality
from ..transformations.safening import no_partial_functions


class CPM_ortools(SolverInterface):
    """
    Interface to OR-Tools' CP-SAT Python API. 

    Creates the following attributes (see parent constructor for more):

    - ``ort_model``: the ortools.sat.python.cp_model.CpModel() created by _model()
    - ``ort_solver``: the ortools cp_model.CpSolver() instance used in solve()

    The :class:`~cpmpy.expressions.globalconstraints.DirectConstraint`, when used, calls a function on the ``ort_model`` object.

    Documentation of the solver's own Python API:
    https://developers.google.com/optimization/reference/python/sat/python/cp_model
    """

    @staticmethod
    def supported():
        # try to import the package
        try:
            import ortools
            return True
        except ModuleNotFoundError:
            return False
        except Exception as e:
            raise e
        
    @staticmethod
    def version() -> Optional[str]:
        """
        Returns the installed version of the solver's Python API.
        """
        try:
            return pkg_resources.get_distribution('ortools').version
        except pkg_resources.DistributionNotFound:
            return None


    def __init__(self, cpm_model=None, subsolver=None):
        """
        Constructor of the native solver object

        Requires a CPMpy model as input, and will create the corresponding
        or-tools model and solver object (ort_model and ort_solver)

        ort_model and ort_solver can both be modified externally before
        calling solve(), a prime way to use more advanced solver features

        Arguments:
            cpm_model: Model(), a CPMpy Model() (optional)
            subsolver: None, not used
        """
        if not self.supported():
            raise Exception("CPM_ortools: Install the python package 'ortools' to use this solver interface.")

        from ortools.sat.python import cp_model as ort

        assert(subsolver is None)

        # initialise the native solver objects
        self.ort_model = ort.CpModel()
        self.ort_solver = ort.CpSolver()

        # for solving with assumption variables,
        # need to store mapping from ORTools Index to CPMpy variable
        self.assumption_dict = None

        # initialise everything else and post the constraints/objective
        super().__init__(name="ortools", cpm_model=cpm_model)

    @property
    def native_model(self):
        """
            Returns the solver's underlying native model (for direct solver access).
        """
        return self.ort_model


    def solve(self, time_limit=None, assumptions=None, solution_callback=None, **kwargs):
        """
            Call the CP-SAT solver

            Arguments:
                time_limit (float, optional):  maximum solve time in seconds 
                assumptions:    list of CPMpy Boolean variables (or their negation) that are assumed to be true.
                                For repeated solving, and/or for use with :func:`s.get_core() <get_core()>`: if the model is UNSAT,
                                get_core() returns a small subset of assumption variables that are unsat together.
                                Note: the or-tools interface is stateless, so you can incrementally call solve() with assumptions, but or-tools will always start from scratch...
                solution_callback (an `ort.CpSolverSolutionCallback` object):   CPMpy includes its own, namely `OrtSolutionCounter`. If you want to count all solutions, 
                                                                                don't forget to also add the keyword argument 'enumerate_all_solutions=True'.
                
                
            The ortools solver parameters are defined in its 'sat_parameters.proto' description:
            https://github.com/google/or-tools/blob/stable/ortools/sat/sat_parameters.proto

            You can use any of these parameters as keyword argument to `solve()` and they will
            be forwarded to the solver. Examples include:

            =============================   ============
            Argument                        Description
            =============================   ============
            ``num_search_workers=8``          number of parallel workers (default: 8)
            ``log_search_progress=True``      to log the search process to stdout (default: False)
            ``cp_model_presolve=False``       to disable presolve (default: True, almost always beneficial)
            ``cp_model_probing_level=0``      to disable probing (default: 2, also valid: 1, maybe 3, etc...)
            ``linearization_level=0``         to disable linearisation (default: 1, can also set to 2)
            ``optimize_with_core=True``       to do max-sat like lowerbound optimisation (default: False)
            ``use_branching_in_lp=True``      to generate more info in lp propagator (default: False)
            ``polish_lp_solution=True``       to spend time in lp propagator searching integer values (default: False)
            ``symmetry_level=1``              only do symmetry breaking in presolve (default: 2, also possible: 0)
            =============================   ============
           

            Examples:

                .. code-block:: python
                
                    o.solve(num_search_workers=8, log_search_progress=True)

        """
        from ortools.sat.python import cp_model as ort
        # ensure all user vars are known to solver
        self.solver_vars(list(self.user_vars))

        # set time limit
        if time_limit is not None:
            if time_limit <= 0:
                raise ValueError("Time limit must be positive")
            self.ort_solver.parameters.max_time_in_seconds = float(time_limit)

        if assumptions is not None:
            ort_assum_vars = self.solver_vars(assumptions)
            # dict mapping ortools vars to CPMpy vars
            self.assumption_dict = {ort_var.Index(): cpm_var for (cpm_var, ort_var) in zip(assumptions, ort_assum_vars)}
            self.ort_model.ClearAssumptions()  # because add just appends
            self.ort_model.AddAssumptions(ort_assum_vars)
            # workaround for a presolve with assumptions bug in ortools
            # https://github.com/google/or-tools/issues/2649
            # still present in v9.0
            self.ort_solver.parameters.keep_all_feasible_solutions_in_presolve = True

        # set additional keyword arguments in sat_parameters.proto
        for (kw, val) in kwargs.items():
            setattr(self.ort_solver.parameters, kw, val)

        if 'log_search_progress' in kwargs and hasattr(self.ort_solver, "log_callback") \
                and (sys.stdout != sys.__stdout__):
            # ortools>9.0, for IPython use, force output redirecting
            # see https://github.com/google/or-tools/issues/1903
            # but only if a nonstandard stdout, otherwise duplicate output
            # see https://github.com/CPMpy/cpmpy/issues/84
            self.ort_solver.log_callback = print

        # call the solver, with parameters
        self.ort_status = self.ort_solver.Solve(self.ort_model, solution_callback=solution_callback)

        # new status, translate runtime
        self.cpm_status = SolverStatus(self.name)
        self.cpm_status.runtime = self.ort_solver.WallTime()

        # translate exit status
        if self.ort_status == ort.FEASIBLE:
            self.cpm_status.exitstatus = ExitStatus.FEASIBLE
        elif self.ort_status == ort.OPTIMAL:
            # COP
            if self.has_objective():
                self.cpm_status.exitstatus = ExitStatus.OPTIMAL
            # CSP
            else:
                # for .solveAll(), if all solutions found status is OPTIMAL
                if kwargs.get("enumerate_all_solutions", False):
                    self.cpm_status.exitstatus = ExitStatus.OPTIMAL
                # regular .solve()
                else:
                    self.cpm_status.exitstatus = ExitStatus.FEASIBLE
        elif self.ort_status == ort.INFEASIBLE:
            self.cpm_status.exitstatus = ExitStatus.UNSATISFIABLE
        elif self.ort_status == ort.MODEL_INVALID:
            raise Exception("OR-Tools says: model invalid:", self.ort_model.Validate())
        elif self.ort_status == ort.UNKNOWN:
            # can happen when timeout is reached...
            self.cpm_status.exitstatus = ExitStatus.UNKNOWN
        else:  # another?
            raise NotImplementedError(self.ort_status)  # a new status type was introduced, please report on github

        # True/False depending on self.cpm_status
        has_sol = self._solve_return(self.cpm_status)

        # translate solution values (of user specified variables only)
        self.objective_value_ = None
        if has_sol:
            # fill in variable values
            for cpm_var in self.user_vars:
                try:
                    cpm_var._value = self.ort_solver.Value(self.solver_var(cpm_var))
                    if isinstance(cpm_var, _BoolVarImpl):
                        cpm_var._value = bool(cpm_var._value) # ort value is always an int
                except IndexError:
                    raise ValueError(f"Var {cpm_var} is unknown to the OR-Tools solver, this is unexpected - "
                                     f"please report on github...")

            # translate objective
            if self.has_objective():
                ort_obj_val = self.ort_solver.ObjectiveValue()
                if round(ort_obj_val) == ort_obj_val: # it is an integer?
                    self.objective_value_ = int(ort_obj_val)  # ensure it is an integer
                else: # can happen when using floats as coeff in objective
                    self.objective_value_ = float(ort_obj_val)
        else: # clear values of variables
            for cpm_var in self.user_vars:
                cpm_var._value = None
        return has_sol

    def solveAll(self, display=None, time_limit=None, solution_limit=None, call_from_model=False, **kwargs):
        """
            A shorthand to (efficiently) compute all solutions, map them to CPMpy and optionally display the solutions.

            It is just a wrapper around the use of `OrtSolutionPrinter()` in fact.

            Arguments:
                display: either a list of CPMpy expressions, OR a callback function, called with the variables after value-mapping. 
                        default/None: nothing displayed
                solution_limit: stop after this many solutions (default: None)
                call_from_model: whether the method is called from a CPMpy Model instance or not

            Returns: 
                number of solutions found
        """
        if self.has_objective():
            raise NotSupportedError("OR-tools does not support finding all optimal solutions.")

        cb = OrtSolutionPrinter(self, display=display, solution_limit=solution_limit)
        self.solve(enumerate_all_solutions=True, solution_callback=cb, time_limit=time_limit, **kwargs)
        return cb.solution_count()


    def solver_var(self, cpm_var):
        """
            Creates solver variable for cpmpy variable
            or returns from cache if previously created
        """
        if is_num(cpm_var):  # shortcut, eases posting constraints
            return cpm_var

        # special case, negative-bool-view
        # work directly on var inside the view
        if isinstance(cpm_var, NegBoolView):
            return self.solver_var(cpm_var._bv).Not()

        # create if it does not exist
        if cpm_var not in self._varmap:
            if isinstance(cpm_var, _BoolVarImpl):
                revar = self.ort_model.NewBoolVar(str(cpm_var))
            elif isinstance(cpm_var, _IntVarImpl):
                revar = self.ort_model.NewIntVar(cpm_var.lb, cpm_var.ub, str(cpm_var))
            else:
                raise NotImplementedError("Not a known var {}".format(cpm_var))
            self._varmap[cpm_var] = revar

        return self._varmap[cpm_var]


    def objective(self, expr, minimize):
        """
            Post the given expression to the solver as objective to minimize/maximize

            - expr: Expression, the CPMpy expression that represents the objective function
            - minimize: Bool, whether it is a minimization problem (True) or maximization problem (False)

            'objective()' can be called multiple times, only the last one is stored

            .. note::
                technical side note: any constraints created during conversion of the objective
                are premanently posted to the solver
        """
        # make objective function non-nested
        (flat_obj, flat_cons) = flatten_objective(expr, csemap=self._csemap)
        self += flat_cons  # add potentially created constraints
        get_variables(flat_obj, collect=self.user_vars)  # add objvars to vars

        # make objective function or variable and post
        obj = self._make_numexpr(flat_obj)
        if minimize:
            self.ort_model.Minimize(obj)
        else:
            self.ort_model.Maximize(obj)

    def has_objective(self):
        return self.ort_model.HasObjective()

    def _make_numexpr(self, cpm_expr):
        """
            Turns a numeric CPMpy 'flat' expression into a solver-specific
            numeric expression

            Used especially to post an expression as objective function

            Accepted by OR-Tools:
            - Decision variable: Var
            - Linear: sum([Var])                                   (CPMpy class 'Operator', name 'sum')
                      wsum([Const],[Var])                          (CPMpy class 'Operator', name 'wsum')
        """
        if is_num(cpm_expr):
            return cpm_expr

        # decision variables, check in varmap
        if isinstance(cpm_expr, _NumVarImpl):  # _BoolVarImpl is subclass of _NumVarImpl
            return self.solver_var(cpm_expr)

        # sum or weighted sum
        if isinstance(cpm_expr, Operator):
            if cpm_expr.name == 'sum':
                return ort.LinearExpr.sum(self.solver_vars(cpm_expr.args))
            elif cpm_expr.name == "sub":
                a,b = self.solver_vars(cpm_expr.args)
                return a - b
            elif cpm_expr.name == 'wsum':
                w = cpm_expr.args[0]
                x = self.solver_vars(cpm_expr.args[1])
                return ort.LinearExpr.weighted_sum(x,w)

        raise NotImplementedError("ORTools: Not a known supported numexpr {}".format(cpm_expr))


    def transform(self, cpm_expr):
        """
            Transform arbitrary CPMpy expressions to constraints the solver supports

            Implemented through chaining multiple solver-independent **transformation functions** from
            the `cpmpy/transformations/` directory.

            See the :ref:`Adding a new solver` docs on readthedocs for more information.

            :param cpm_expr: CPMpy expression, or list thereof
            :type cpm_expr: Expression or list of Expression

            :return: list of Expression
        """
        cpm_cons = toplevel_list(cpm_expr)
        supported = {"min", "max", "abs", "element", "alldifferent", "xor", "table", "negative_table", "cumulative", "circuit", "inverse", "no_overlap", "regular"}
        cpm_cons = no_partial_functions(cpm_cons, safen_toplevel=frozenset({"div", "mod"})) # before decompose, assumes total decomposition for partial functions
        cpm_cons = decompose_in_tree(cpm_cons, supported, csemap=self._csemap)
        cpm_cons = flatten_constraint(cpm_cons, csemap=self._csemap)  # flat normal form
        cpm_cons = reify_rewrite(cpm_cons, supported=frozenset(['sum', 'wsum']), csemap=self._csemap)  # constraints that support reification
        cpm_cons = only_numexpr_equality(cpm_cons, supported=frozenset(["sum", "wsum", "sub"]), csemap=self._csemap)  # supports >, <, !=
        cpm_cons = only_bv_reifies(cpm_cons, csemap=self._csemap)
        cpm_cons = only_implies(cpm_cons, csemap=self._csemap)  # everything that can create
                                             # reified expr must go before this
        return cpm_cons

    def add(self, cpm_expr):
        """
            Eagerly add a constraint to the underlying solver.

            Any CPMpy expression given is immediately transformed (through `transform()`)
            and then posted to the solver in this function.

            This can raise 'NotImplementedError' for any constraint not supported after transformation

            The variables used in expressions given to add are stored as 'user variables'. Those are the only ones
            the user knows and cares about (and will be populated with a value after solve). All other variables
            are auxiliary variables created by transformations.

            :param cpm_expr: CPMpy expression, or list thereof
            :type cpm_expr: Expression or list of Expression

            :return: self
        """
        # add new user vars to the set
        get_variables(cpm_expr, collect=self.user_vars)

        # transform and post the constraints
        for con in self.transform(cpm_expr):
            self._post_constraint(con)

        return self
    __add__ = add  # avoid redirect in superclass

    # TODO: 'reifiable' is an artefact from the early days
    # only 3 constraints support it (and,or,sum),
    # we can just add reified support for those and not need `reifiable` or returning the constraint
    # then we can remove _post_constraint and have its code inside the for loop of `add()`
    # like for other solvers
    def _post_constraint(self, cpm_expr, reifiable=False):
        """
            Post a supported CPMpy constraint directly to the underlying solver's API

            What 'supported' means depends on the solver capabilities, and in effect on what transformations
            are applied in `transform()`.

            Returns the posted ortools 'Constraint', so that it can be used in reification
            e.g. self._post_constraint(smth, reifiable=True).onlyEnforceIf(self.solver_var(bvar))

        :param cpm_expr: CPMpy expression
        :type cpm_expr: Expression

        :param reifiable: if True, will throw an error if cpm_expr can not be reified by ortools (for safety)
        """

        # Operators: base (bool), lhs=numexpr, lhs|rhs=boolexpr (reified ->)
        if isinstance(cpm_expr, Operator):
            # 'and'/n, 'or'/n, '->'/2
            if cpm_expr.name == 'and':
                return self.ort_model.AddBoolAnd(self.solver_vars(cpm_expr.args))
            elif cpm_expr.name == 'or':
                return self.ort_model.AddBoolOr(self.solver_vars(cpm_expr.args))
            elif cpm_expr.name == '->':
                assert(isinstance(cpm_expr.args[0], _BoolVarImpl))  # lhs must be boolvar
                lhs = self.solver_var(cpm_expr.args[0])
                if isinstance(cpm_expr.args[1], _BoolVarImpl):
                    # bv -> bv
                    return self.ort_model.AddImplication(lhs, self.solver_var(cpm_expr.args[1]))
                else:
                    # bv -> boolexpr
                    # the `reify_rewrite()` transformation ensures that only
                    # the natively reifiable 'and', 'or' and 'sum' remain here
                    return self._post_constraint(cpm_expr.args[1], reifiable=True).OnlyEnforceIf(lhs)
            else:
                raise NotImplementedError("Not a known supported ORTools Operator '{}' {}".format(
                        cpm_expr.name, cpm_expr))

        # Comparisons: only numeric ones as the `only_implies()` transformation
        # has removed the '==' reification for Boolean expressions
        # numexpr `comp` bvar|const
        elif isinstance(cpm_expr, Comparison):
            lhs = cpm_expr.args[0]
            ortrhs = self.solver_var(cpm_expr.args[1])

            if isinstance(lhs, _NumVarImpl):
                # both are variables, do python comparison over ORT variables
                return self.ort_model.Add(eval_comparison(cpm_expr.name, self.solver_var(lhs), ortrhs))
            elif isinstance(lhs, Operator) and (lhs.name == 'sum' or lhs.name == 'wsum' or lhs.name == "sub"):
                # a BoundedLinearExpression LHS, special case, like in objective
                ortlhs = self._make_numexpr(lhs)
                # ortools accepts sum(x) >= y over ORT variables
                return self.ort_model.Add(eval_comparison(cpm_expr.name, ortlhs, ortrhs))
            elif cpm_expr.name == '==':
                # NumExpr == IV, supported by ortools (thanks to `only_numexpr_equality()` transformation)
                if lhs.name == 'min':
                    return self.ort_model.AddMinEquality(ortrhs, self.solver_vars(lhs.args))
                elif lhs.name == 'max':
                    return self.ort_model.AddMaxEquality(ortrhs, self.solver_vars(lhs.args))
                elif lhs.name == 'abs':
                    return self.ort_model.AddAbsEquality(ortrhs, self.solver_var(lhs.args[0]))
                elif lhs.name == 'mul':
                    return self.ort_model.AddMultiplicationEquality(ortrhs, self.solver_vars(lhs.args))
                elif lhs.name == 'div':
                    return self.ort_model.AddDivisionEquality(ortrhs, *self.solver_vars(lhs.args))
                elif lhs.name == 'element':
                    arr, idx = lhs.args
                    if is_int(idx): # OR-Tools does not handle all constant integer cases
                        idx = intvar(idx,idx)
                    # OR-Tools has slight different in argument order
                    return self.ort_model.AddElement(
                        self.solver_var(idx),
                        self.solver_vars(arr),
                        ortrhs
                    )
                elif lhs.name == 'mod':
                    # catch tricky-to-find ortools limitation
                    x,y = lhs.args
                    if get_bounds(y)[0] <= 0: # not supported, but result of modulo is agnositic to sign of second arg
                        y, link = get_or_make_var(-lhs.args[1], csemap=self._csemap)
                        self += link
                    return self.ort_model.AddModuloEquality(ortrhs, *self.solver_vars([x,y]))
                elif lhs.name == 'pow':
                    # only `POW(b,2) == IV` supported, post as b*b == IV
                    if not is_num(lhs.args[1]):
                        raise NotSupportedError(f"OR-Tools does not support power constraints with variable as exponent, got {lhs}")
                    if lhs.args[1] == 2:
                        b = self.solver_var(lhs.args[0])
                        return self.ort_model.AddMultiplicationEquality(ortrhs, [b,b])
                    else: # need to create intermediate vars, post (((b * b) * b) * ...) * b == IV
                        b, n = lhs.args
                        new_lhs = 1
                        for exp in range(n):
                            new_lhs, new_cons = get_or_make_var(b * new_lhs, csemap=self._csemap)
                            self += new_cons
                        return self.ort_model.Add(eval_comparison("==", self.solver_var(new_lhs), ortrhs))


            raise NotImplementedError(
                        "Not a known supported ORTools left-hand-side '{}' {}".format(lhs.name, cpm_expr))

        # base (Boolean) global constraints
        elif isinstance(cpm_expr, GlobalConstraint):

            if cpm_expr.name == 'alldifferent':
                return self.ort_model.AddAllDifferent(self.solver_vars(cpm_expr.args))
            elif cpm_expr.name == 'table':
                assert (len(cpm_expr.args) == 2)  # args = [array, table]
                array, table = cpm_expr.args
                array = self.solver_vars(array)
                # table needs to be a list of lists of integers
                return self.ort_model.AddAllowedAssignments(array, table)
            elif cpm_expr.name == 'negative_table':
                assert (len(cpm_expr.args) == 2)  # args = [array, table]
                array, table = cpm_expr.args
                array = self.solver_vars(array)
                return self.ort_model.AddForbiddenAssignments(array, table)
            elif cpm_expr.name == "regular":
                array, transitions, start, accepting = cpm_expr.args
                array = self.solver_vars(array)
                return self.ort_model.AddAutomaton(array, cpm_expr.node_map[start], [cpm_expr.node_map[n] for n in accepting], 
                                                   [(cpm_expr.node_map[src], label, cpm_expr.node_map[dst]) for src, label, dst in transitions])
            elif cpm_expr.name == "cumulative":
                start, dur, end, demand, cap = self.solver_vars(cpm_expr.args)
                intervals = [self.ort_model.NewIntervalVar(s,d,e,f"interval_{s}-{d}-{e}") for s,d,e in zip(start,dur,end)]
                return self.ort_model.AddCumulative(intervals, demand, cap)
            elif cpm_expr.name == "no_overlap":
                start, dur, end = self.solver_vars(cpm_expr.args)
                intervals = [self.ort_model.NewIntervalVar(s,d,e, f"interval_{s}-{d}-{d}") for s,d,e in zip(start,dur,end)]
                return self.ort_model.add_no_overlap(intervals)
            elif cpm_expr.name == "circuit":
                # ortools has a constraint over the arcs, so we need to create these
                # when using an objective over arcs, using these vars direclty is recommended
                # (see PCTSP-path model in the future)
                x = cpm_expr.args
                N = len(x)
                arcvars = boolvar(shape=(N,N))
                # post channeling constraints from int to bool
                self += [b == (x[i] == j) for (i,j),b in np.ndenumerate(arcvars)]
                # post the global constraint
                # when posting arcs on diagonal (i==j), it would do subcircuit
                ort_arcs = [(i,j,self.solver_var(b)) for (i,j),b in np.ndenumerate(arcvars) if i != j]
                return self.ort_model.AddCircuit(ort_arcs)
            elif cpm_expr.name == 'inverse':
                assert len(cpm_expr.args) == 2, "inverse() expects two args: fwd, rev"
                fwd, rev = self.solver_vars(cpm_expr.args)
                return self.ort_model.AddInverse(fwd, rev)
            elif cpm_expr.name == 'xor':
                args = cpm_expr.args
                if any(is_true_cst(a) for a in cpm_expr.args):
                    # replace with constant variable instead
                    if not hasattr(self, "_true_var"):
                        self._true_var = boolvar()
                        self.add(self._true_var)
                    args = [a if not is_true_cst(a) else self._true_var for a in cpm_expr.args]

                # remove false constants
                args = [a for a in args if not is_false_cst(a)]
                return self.ort_model.AddBoolXOr(self.solver_vars(args))
            else:
                raise NotImplementedError(f"Unknown global constraint {cpm_expr}, should be decomposed! "
                                          f"If you reach this, please report on github.")

        # unlikely base case: Boolean variable
        elif isinstance(cpm_expr, _BoolVarImpl):
            return self.ort_model.AddBoolOr([self.solver_var(cpm_expr)])

        # unlikely base case: True or False
        elif isinstance(cpm_expr, BoolVal):
            return self.ort_model.Add(cpm_expr.args[0])

        # a direct constraint, pass to solver
        elif isinstance(cpm_expr, DirectConstraint):
            return cpm_expr.callSolver(self, self.ort_model)

        # else
        raise NotImplementedError(cpm_expr)  # if you reach this... please report on github


    def solution_hint(self, cpm_vars, vals):
        """
        OR-Tools supports warmstarting the solver with a feasible solution.

        More specifically, it will branch that variable on that value first if possible. This is known as 'phase saving' in the SAT literature, but then extended to integer variables.

        The solution hint does NOT need to satisfy all constraints, it should just provide reasonable default values for the variables. It can decrease solving times substantially, especially when solving a similar model repeatedly.

        :param cpm_vars: list of CPMpy variables
        :param vals: list of (corresponding) values for the variables
        """
        self.ort_model.ClearHints() # because add just appends

        cpm_vars = flatlist(cpm_vars)
        vals = flatlist(vals)
        assert (len(cpm_vars) == len(vals)), "Variables and values must have the same size for hinting"
        for (cpm_var, val) in zip(cpm_vars, vals):
            self.ort_model.AddHint(self.solver_var(cpm_var), val)


    def get_core(self):
        """
            For use with :func:`s.solve(assumptions=[...]) <solve()>`. Only meaningful if the solver returned UNSAT. In that case, ``get_core()`` returns a small subset of assumption variables that are unsat together.

            CPMpy will return only those variables that are False (in the UNSAT core)

            Note that there is no guarantee that the core is minimal, though this interface does open up the possibility to add more advanced Minimal Unsatisfiabile Subset algorithms on top. All contributions welcome!

            For pure OR-Tools example, see http://github.com/google/or-tools/blob/master/ortools/sat/samples/assumptions_sample_sat.py

            Requires ortools >= 8.2!!!
        """
        from ortools.sat.python import cp_model as ort
        assert (self.ort_status == ort.INFEASIBLE), "get_core(): solver must return UNSAT"
        assert (self.assumption_dict is not None),  "get_core(): requires a list of assumption variables, e.g. s.solve(assumptions=[...])"

        # use our own dict because of VarIndexToVarProto(0) bug in ort 8.2
        assum_idx = self.ort_solver.SufficientAssumptionsForInfeasibility()

        # return cpm_variables corresponding to ort_assum vars in UNSAT core
        return [self.assumption_dict[i] for i in assum_idx]

    @classmethod
    def tunable_params(cls):
        """
            Suggestion of tunable hyperparameters of the solver.
            List compiled based on a conversation with OR-tools' Laurent Perron (issue #138).
        """
        return {
            # 'use_branching_in_lp': [False, True], # removed in OR-tools >= v9.9
            'optimize_with_core' : [False, True],
            'search_branching': [0,1,2,3,4,5,6],
            'boolean_encoding_level' : [0,1,2,3],
            'linearization_level': [0, 1, 2],
            'core_minimization_level' : [0,1,2], # new in OR-tools>= v9.8
            'cp_model_probing_level': [0, 1, 2, 3],
            'cp_model_presolve' : [False, True],
            'clause_cleanup_ordering' : [0,1],
            'binary_minimization_algorithm' : [0,1,2,3,4],
            'minimization_algorithm' : [0,1,2,3],
            'use_phase_saving' : [False, True]
        }

    @classmethod
    def default_params(cls):
        return {
            # 'use_branching_in_lp': False, # removed in OR-tools >= v9.9
            'optimize_with_core': False,
            'search_branching': 0,
            'boolean_encoding_level': 1,
            'linearization_level': 1,
            'core_minimization_level': 2,# new in OR-tools>=v9.8
            'cp_model_probing_level': 2,
            'cp_model_presolve': True,
            'clause_cleanup_ordering': 0,
            'binary_minimization_algorithm': 1,
            'minimization_algorithm': 2,
            'use_phase_saving': True
        }


# solvers are optional, so this file should be interpretable
# even if ortools is not installed...
try:
    from ortools.sat.python import cp_model as ort
    import time


    class OrtSolutionCounter(ort.CpSolverSolutionCallback):
        """
        Native ortools callback for solution counting.

        It is based on ortools' built-in `ObjectiveSolutionPrinter`
        but with output printing being optional

        use with CPM_ortools as follows:

        .. code-block:: python
            
            cb = OrtSolutionCounter()
            s.solve(enumerate_all_solutions=True, solution_callback=cb)

        then retrieve the solution count with ``cb.solution_count()``

        Arguments:
            verbose (bool, default: False): whether to print info on every solution found 
    """

        def __init__(self, verbose=False):
            super().__init__()
            self.__solution_count = 0
            self.__verbose = verbose
            if self.__verbose:
                self.__start_time = time.time()

        def on_solution_callback(self):
            """Called on each new solution."""
            if self.__verbose:
                current_time = time.time()
                obj = self.ObjectiveValue()
                print('Solution %i, time = %0.2f s, objective = %i' %
                      (self.__solution_count, current_time - self.__start_time, obj))
            self.__solution_count += 1

        def solution_count(self):
            """Returns the number of solutions found."""
            return self.__solution_count

    class OrtSolutionPrinter(OrtSolutionCounter):
        """
            Native OR-Tools callback for solution printing.

            Subclasses :class:`OrtSolutionCounter`, see those docs too.

            Use with :class:`CPM_ortools` as follows:

            .. code-block:: python

                cb = OrtSolutionPrinter(s, display=vars)
                s.solve(enumerate_all_solutions=True, solution_callback=cb)

            For multiple variables (single or NDVarArray), use:
            ``cb = OrtSolutionPrinter(s, display=[v, x, z])``.

            For a custom print function, use for example:
            
            .. code-block:: python

                def myprint():
                    print(f"x0={x[0].value()}, x1={x[1].value()}")
                    cb = OrtSolutionPrinter(s, printer=myprint)

            Optionally retrieve the solution count with ``cb.solution_count()``.

            Arguments:
                verbose (bool, default = False): whether to print info on every solution found 
                display: either a list of CPMpy expressions, OR a callback function, called with the variables after value-mapping
                            default/None: nothing displayed
                solution_limit (default = None): stop after this many solutions 
        """
        def __init__(self, solver, display=None, solution_limit=None, verbose=False):
            super().__init__(verbose)
            self._solution_limit = solution_limit
            # we only need the cpmpy->solver varmap from the solver
            self._varmap = solver._varmap
            # identify which variables to populate with their values
            self._cpm_vars = []
            self._display = display
            if isinstance(display, (list,Expression)):
                self._cpm_vars = get_variables(display)
            elif callable(display):
                # might use any, so populate all (user) variables with their values
                self._cpm_vars = solver.user_vars

        def on_solution_callback(self):
            """Called on each new solution."""
            super().on_solution_callback()
            if len(self._cpm_vars):
                # populate values before printing
                for cpm_var in self._cpm_vars:
                    # it might be an NDVarArray
                    if hasattr(cpm_var, "flat"):
                        for cpm_subvar in cpm_var.flat:
                            cpm_subvar._value = self.Value(self._varmap[cpm_subvar])
                    elif isinstance(cpm_var, _BoolVarImpl):
                        cpm_var._value = bool(self.Value(self._varmap[cpm_var]))
                    else:
                        cpm_var._value = self.Value(self._varmap[cpm_var])

                if isinstance(self._display, Expression):
                    print(argval(self._display))
                elif isinstance(self._display, list):
                    # explicit list of expressions to display
                    print(argvals(self._display))
                else: # callable
                    self._display()

            # check for count limit
            if self.solution_count() == self._solution_limit:
                self.StopSearch()

except ImportError:
    pass  # Ok, no ortools installed...
